# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
#cython: cdivision=True
#cython: boundscheck=False
#cython: profile=False

from __future__ import print_function

# C libraries
from cython.parallel cimport prange, parallel
from libc.stdio cimport FILE, fopen, fclose, fseek, ftell, SEEK_SET, fgets, printf, fread
from libc.string cimport strncpy, strlen, strcmp
from libc.stdlib cimport atof, atoi
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from libc.stdlib cimport free, malloc

from .core_base cimport OPENMP_NUM_THREADS, PBCBox, TopologyReader, IndexReader, CoordinateReader
from .typedefs cimport real, rvec, fsl_int, fsl_uint
cimport cython
import numpy as np

cdef extern from "xdrfile/xdrfile.h":
    ctypedef struct XDRFILE:
        pass

    # XDRFile I/O
    XDRFILE *xdrfile_open(const char *path, const char *mode) nogil
    int xdrfile_close(XDRFILE *xfp) nogil
    int xdrfile_getpos(XDRFILE* xfp) nogil
    int xdrfile_setpos(XDRFILE* xfp, unsigned int pos) nogil

    # XDRFile Reading
    int xdrfile_decompress_coord_double_partial(double *ptr,
                                             int *size,
                                             int *ncoords,
                                             double *precision,
                                             XDRFILE* xfp) nogil
    int xdrfile_read_float(float *           ptr,
					       int               ndata,
                           XDRFILE *         xfp) nogil
    int xdrfile_read_double(double *           ptr,
					        int               ndata,
                            XDRFILE *         xfp) nogil
    int xdrfile_read_int(int *         ptr,
					     int           ndata,
					     XDRFILE *     xfp) nogil



# Python libraries
import numpy
import os

# Constants
DEF XTC_MAGIC = 1995
DEF TRR_MAGIC = 1993
DEF TRR_HEADER_SIZE = 76
DEF XTC_HEADER_SIZE = 56
DEF DIM = 3
DEF XX = 0
DEF YY = 1
DEF ZZ = 2
DEF RET_OK = 1
DEF RET_YES = 1
DEF RET_ERROR = 0
DEF RET_NO = 0

ctypedef float fvec[DIM]

cdef bint is_solvent(bytes resname):

    if resname == b"W":
        return True

    if resname == b"PW":
        return True

    if resname == b"SOL":
        return True

    if resname == b"WAT":
        return True

    if resname == b"HOH":
        return True

    if resname == b"OHH":
        return True

    if resname == b"TIP":
        return True

    if resname == b"T3P":
        return True

    if resname == b"T4P":
        return True

    if resname == b"T5P":
        return True

    if resname == b"T3H":
        return True
    return False

cdef class GroReaderTopol(TopologyReader):
    cdef load(self):
        cdef int lino = -1, natoms = -1, last_resid = -1
        cdef int resid, atomid, resid_offset = 0
        cdef bytes resname
        cdef str atomname
        cdef bint skip = False

        topol = self.topology
        with open(self.filename, "r") as fp:
            for line in fp:
                lino += 1
                if lino == 0:
                    continue
                elif lino == 1:
                    natoms = int(line)
                else:
                    if lino < natoms + 2:
                        resid = atoi(line[:5].encode())

                        # Correct the resid if necessary because resids are modulo 10000 in .gro
                        if (resid+resid_offset * 10000) < last_resid:
                            resid_offset += 1
                        resid += resid_offset * 10000

                        if resid == last_resid and skip:
                            if skip: # Already selected to be skipped
                                continue
                        else:
                            if resid != last_resid:
                                skip = False

                            # Update the last_resid counter
                            last_resid = resid

                            # Read residue name
                            resname = line[5:10].strip().encode()

                            if is_solvent(resname): # Classified as solvent, we skip it
                                residue = None
                                skip = True
                                #print("Skipping %s - resid:%i" % (resname, resid))
                                continue

                            # Read atom information
                            atomname = line[10:15].strip()
                            #atomname = strip_name(line[10:15])

                            if atomname[0:1] == "H": # Labelled as hydrogen -> skipped
                                continue

                            atomid = atoi(line[15:20].encode()) + ((lino - 1)// 100000) * 100000

                            topol.internal_append(resid, resname, atomname.encode(), atomid)




_Topology_readers = {".gro": GroReaderTopol}

cdef class NdxReader(IndexReader):
    cdef fast_load(self):
        cdef str line
        cdef bytes last_group = b""

        if self.filename.split(b".")[-1] != b"ndx":
            raise ValueError("%s is not a supported index file. supported: .ndx" % self.filename)
        groups = {}

        with open(self.filename, "r") as fp:

            # Use Python object to parse line... Much more easier!
            for line in fp:
                if line[0] == '[':
                    if last_group != b"":
                        groups[last_group] = numpy.fromstring(groups[last_group], dtype=np.int64, sep=" ")
                    last_group = line.strip(" []\n").encode()
                    groups[last_group] = b""
                else:
                    groups[last_group] += line.strip().encode() + b" "

        groups[last_group] = numpy.fromstring(groups[last_group], dtype=np.int64, sep=" ")

        self.groups = groups

_Index_loaders = {".ndx": NdxReader}


cdef class GroReaderCoords(CoordinateReader):
    cdef fsl_int coordline_length

    cdef preload(self):
        cdef int lino = -1, natoms = -1
        cdef bint need_read = True

        self.nframes = 1
        self.timesteps = numpy.array([0], dtype=np.float64)

        if self.filename.decode().split(".")[-1] != "gro":
            raise ValueError("%s is not a supported gro file. supported: .gro" % self.filename)

        with open(self.filename, "r") as fp:
            while need_read:
                lino += 1
                line = fp.readline()
                if lino == 0:
                    continue
                elif lino == 1:
                    natoms = int(line)
                    self.coordinate_offsets = numpy.array((fp.tell(),), dtype=np.int64)
                elif lino == 2:
                    self.coordline_length = len(line)
                elif lino == (natoms + 1):
                    self.box_offsets = numpy.array((fp.tell(),), dtype=np.int64)
                    need_read = False



    cdef PBCBox load_box(self, int frame_id):
        cdef int n_coords, i, j
        self.assert_frame_id(frame_id)

        box = numpy.zeros((3,3), np.float64)

        with open(self.filename, "rb") as fp:
            with cython.boundscheck(True):
                fp.seek(self.box_offsets[frame_id])
            box_str = fp.read().strip(b'\n\r').split()

        n_coords = len(box_str)

        box[0, 0] = float(box_str[0])
        box[1, 1] = float(box_str[1])
        box[2, 2] = float(box_str[2])

        if n_coords == 9:
            box[0, 1] = float(box_str[3])
            box[0, 2] = float(box_str[4])
            box[1, 0] = float(box_str[5])
            box[1, 2] = float(box_str[6])
            box[2, 0] = float(box_str[7])
            box[2, 1] = float(box_str[8])

        return PBCBox(box)

    cdef real[:,::1] load_coords(self, int frame_id, fsl_int[:] atomids) nogil except *:
        cdef int natoms = atomids.shape[0]
        cdef int i
        cdef FILE* cfile
        cdef real[:,::1] coords_memview
        cdef char* cfilename
        cdef fsl_int coords_offset
        cdef fsl_int coord_offset
        cdef int min_offset
        cdef int max_offset
        cdef int offset_range
        cdef char *coords_buffer
        cdef char *coord_buffer
        cdef double val

        with gil:
            coords_memview = numpy.empty((natoms, 3))
            cfilename = self.filename
            coords_offset = self.coordinate_offsets[frame_id]

        # Do nothing if there is no atom to read!
        if natoms == 0:
             return coords_memview


        min_offset = atomids[0]
        max_offset = atomids[0]
        for i in range(natoms):
            if atomids[i] < min_offset:
                min_offset = atomids[i]
            elif atomids[i] > max_offset:
                max_offset = atomids[i]

        offset_range = (max_offset - min_offset + 1) * self.coordline_length # +1 is because we also the last line!


        with gil:
            coords_buffer = <char *> PyMem_Malloc(offset_range * sizeof(char))

        # Open coords file
        cfile = fopen(cfilename, "rb")
        if cfile == NULL:
            with gil:
                raise IOError("No such file or directory: '%s'" % self.filename)

        # Seek the proper position
        fseek(cfile, coords_offset + (min_offset - 1) * self.coordline_length, SEEK_SET) # NOTE: atomid starts @ 1

        # Read buffer & close the file
        if fread(coords_buffer, offset_range, 1, cfile) != 1:
            fclose(cfile)
            with gil:
                raise RuntimeError("EOF reached!")
        fclose(cfile)

        with parallel(num_threads=OPENMP_NUM_THREADS):
            coord_buffer = <char *> malloc(9 * sizeof(char))
            # Initialize buffer end!
            coord_buffer[8] = b'\0'

            for i in prange(natoms, schedule='static'):
                coord_offset = (atomids[i] - min_offset) * self.coordline_length

                # Extract coordinates from buffer
                strncpy(coord_buffer, coords_buffer+coord_offset+20, 8)
                coords_memview[i, XX] = atof(coord_buffer)

                strncpy(coord_buffer, coords_buffer+coord_offset+28, 8)
                coords_memview[i, YY] = atof(coord_buffer)

                strncpy(coord_buffer, coords_buffer+coord_offset+36, 8)
                coords_memview[i, ZZ] = atof(coord_buffer)

            free(coord_buffer)

        with gil:
            PyMem_Free(coords_buffer)

        return coords_memview


cdef class XtcReaderCoords(CoordinateReader):
    cdef preload(self):
        cdef fsl_uint offset=0
        cdef fsl_uint max_offset = os.path.getsize(self.filename)
        cdef int magic=-1, num_atoms_frame, frame_size, num_coords_bytes, nframes = 0
        cdef float timestep_float
        cdef XDRFILE *xfp

        frame_offsets = []
        timesteps = []

        # Open trajectory
        xfp = xdrfile_open(self.filename, "rb")
        if xfp == NULL:
            raise IOError("Could not read file: %s" % self.filename)

        # Loop over all available frames
        while offset < max_offset:
            # Get to the proper offset
            if xdrfile_setpos(xfp, offset) == 0:
                raise IOError("Could not set position in file: %s" % self.filename)

            # Initialize the frame length
            frame_size = XTC_HEADER_SIZE

            # Check the magic number
            xdrfile_read_int(&magic, 1, xfp)
            if magic != XTC_MAGIC:
                raise IOError("Corrupted XTC file: '%s'" % self.filename)

            # Read timestep
            if xdrfile_setpos(xfp, offset + 12) == 0:
                raise IOError("Could not set position in file: %s" % self.filename)
            xdrfile_read_float(&timestep_float, 1, xfp)

            # Read number of atoms
            if xdrfile_setpos(xfp, offset + 52) == 0:
                raise IOError("Could not set position in file: %s" % self.filename)
            xdrfile_read_int(&num_atoms_frame, 1, xfp)

            if num_atoms_frame <= 9: # No compression used
                frame_size += num_atoms_frame * 3 * 4
            else:
                # Read number of bytes used to store coordinates
                if xdrfile_setpos(xfp, offset + 88) == 0:
                    raise IOError("Could not set position in file: %s" % self.filename)
                xdrfile_read_int(&num_coords_bytes, 1, xfp)

                # As coords bytes might not a factor of 4, we need to make sure of that.
                # NOTE: this is due to xdr file spec, not related to gmx stuff.
                if num_coords_bytes % 4 != 0:
                    num_coords_bytes = 4 * (num_coords_bytes // 4 + 1)

                frame_size += 36 + num_coords_bytes

            # Store data
            frame_offsets.append(offset)
            timesteps.append(timestep_float)

            # Update frame counter and offset
            nframes += 1
            offset += frame_size

        # Close file and release memory
        xdrfile_close(xfp)

        frame_offsets = numpy.array(frame_offsets, dtype=np.int64)
        timesteps = numpy.array(timesteps, dtype=np.float64)
        self.timesteps = timesteps
        self.coordinate_offsets = frame_offsets + 52
        self.box_offsets = frame_offsets + 16
        self.nframes = nframes

    cdef PBCBox load_box(self, int frame_id):
        cdef XDRFILE *xfp
        cdef fsl_uint pos = self.box_offsets[frame_id]
        cdef real[:,::1] box = numpy.empty((DIM, DIM), dtype=np.float64)
        cdef float float_box[DIM][DIM]
        cdef int i,j

        xfp = xdrfile_open(self.filename, "rb")
        if xfp == NULL:
            raise IOError("Could not read file: %s" % self.filename)
        if xdrfile_setpos(xfp, pos) == 0:
            raise IOError("Could not set position in file: %s" % self.filename)
        if xdrfile_read_float(float_box[0], DIM*DIM, xfp) != DIM*DIM:
            raise IOError("Could not read matrix")
        # Convert float to real
        for i in range(DIM):
            for j in range(DIM):
                box[i, j] = <real> float_box[i][j]
        # Close file and release memory
        xdrfile_close(xfp)

        return PBCBox(box)



    cdef real[:, ::1] load_coords(self, int frame_id, fsl_int[:] atomids) nogil except *:
        cdef XDRFILE *xfp
        cdef fsl_uint pos
        cdef int size = atomids.shape[0], total_size
        cdef int i
        cdef fsl_int tmpid, maxid = -1
        cdef int natoms, ndecompressed
        cdef real[:, ::1] coords
        cdef rvec *real_coords
        cdef real prec = 0.0
        cdef char *filename

        with gil:
            coords = numpy.empty((size, DIM))
            pos = self.coordinate_offsets[frame_id]
            filename = self.filename

        # Retrieve the max atomid to get how many coordinates needs to be retrieved
        for i in range(size):
            tmpid = atomids[i]
            if tmpid > maxid:
                maxid = tmpid
        natoms = maxid

        # Open traj file
        xfp = xdrfile_open(filename, "rb")
        if xfp == NULL:
            with gil:
                raise IOError("Could not read file: %s" % self.filename)
        if xdrfile_setpos(xfp, pos) == 0:
            with gil:
                raise IOError("Could not set position in file: %s" % self.filename)

        # Get the number of coordinates in frame
        xdrfile_read_int(&total_size, 1, xfp)

        # Allocate memory
        with gil:
            real_coords = <rvec *> PyMem_Malloc(sizeof(rvec) * total_size)

        # Reset position to the beginning of the frame
        if xdrfile_setpos(xfp, pos) == 0:
            with gil:
                raise IOError("Could not set position in file: %s" % self.filename)

        # Load coordinates
        if xdrfile_decompress_coord_double_partial(real_coords[0], &total_size, &natoms, &prec, xfp) < -1:
            with gil:
                raise IOError("Could not decompress coordinates")

        # Close file
        xdrfile_close(xfp)

        # Build the output array
        for i in range(size):
            tmpid = atomids[i] - 1 # Reminder: atomids start at 1

            coords[i, XX] = real_coords[tmpid][XX]
            coords[i, YY] = real_coords[tmpid][YY]
            coords[i, ZZ] = real_coords[tmpid][ZZ]

        # Free memory
        with gil:
            PyMem_Free(real_coords)

        return coords

cdef class TrrReaderCoords(CoordinateReader):
    cdef bint use_double
    cdef int natoms
    cdef preload(self):
        cdef fsl_uint offset=0
        cdef fsl_uint max_offset = os.path.getsize(self.filename)
        cdef int magic=-1, num_atoms_frame, frame_size, num_coords_bytes, nframes = 0
        cdef float timestep_float
        cdef real timestep
        cdef XDRFILE *xfp
        cdef char *version = "GMX_trn_file"
        cdef int ir_size, e_size, box_size, vir_size, pres_size, top_size, sym_size, x_size, v_size, f_size
        cdef int float_size = 0
        cdef bint use_double

        frame_offsets = []
        timesteps = []

        self.natoms = 0

        # Open trajectory
        xfp = xdrfile_open(self.filename, "rb")
        if xfp == NULL:
            raise IOError("Could not read file: %s" % self.filename)

        # Loop over all available frames
        while offset < max_offset:
            # Get to the proper offset
            if xdrfile_setpos(xfp, offset) == 0:
                raise IOError("Could not set position in file: %s" % self.filename)

            # Check the magic number
            xdrfile_read_int(&magic, 1, xfp)
            if magic != TRR_MAGIC:
                raise IOError("Corrupted TRR file: '%s'" % self.filename)

            # Check version string
            xdrfile_read_int(&frame_size, 1, xfp)
            if frame_size != strlen(version)+1:
                raise IOError("Corrupted TRR file: '%s'" % self.filename)

            if xdrfile_setpos(xfp, offset + 8 + 16) == 0: # Skip the version string
                raise IOError("Could not set position in file: %s" % self.filename)
            xdrfile_read_int(&ir_size, 1, xfp)
            xdrfile_read_int(&e_size, 1, xfp)
            xdrfile_read_int(&box_size, 1, xfp)
            xdrfile_read_int(&vir_size, 1, xfp)
            xdrfile_read_int(&pres_size, 1, xfp)
            xdrfile_read_int(&top_size, 1, xfp)
            xdrfile_read_int(&sym_size, 1, xfp)
            xdrfile_read_int(&x_size, 1, xfp)
            xdrfile_read_int(&v_size, 1, xfp)
            xdrfile_read_int (&f_size, 1, xfp)
            xdrfile_read_int(&num_atoms_frame, 1, xfp)

            frame_size = TRR_HEADER_SIZE + ir_size + e_size + box_size + vir_size + pres_size + top_size + sym_size + \
                         x_size + v_size + f_size

            if x_size == 0:
                raise ValueError("Frames without coordinates are not supported. Please correct your trajectory")

            # Check trajectory homogeneity
            if offset == 0:
                self.natoms = num_atoms_frame
            else:
                if num_atoms_frame != self.natoms:
                    raise ValueError("Only trajectory with same number of atoms per frame is supported.\n\
                                     Please correct your trajectory file.")

            # Check float size
            if box_size != 0:
                float_size = box_size / 9
            else:
                float_size = x_size / (3 * num_atoms_frame)

            if float_size == 4:
                use_double = False
            elif float_size == 8:
                use_double = True
            else:
                raise ValueError("Float size is not correct (value: %i), trajectory must be corrupted!" % float_size)

            if offset == 0:
                self.use_double = use_double
            else:
                if self.use_double != use_double:
                    raise ValueError("Only trajectory with homogeneous float size is supported.")

            if xdrfile_setpos(xfp, offset + TRR_HEADER_SIZE) == 0: # Skip the step and nre
                raise IOError("Could not set position in file: %s" % self.filename)
            if use_double:
                xdrfile_read_double(&timestep, 1, xfp)
            else:
                xdrfile_read_float(&timestep_float, 1, xfp)
                timestep = timestep_float

            # Store data
            frame_offsets.append(offset)
            timesteps.append(timestep)

            # Update frame counter and offset
            nframes += 1
            offset += frame_size + 2 * float_size # the 2 floats are timestep and lamba_val


        xdrfile_close(xfp)
        frame_offsets = numpy.array(frame_offsets, dtype=np.int64)
        timesteps = numpy.array(timesteps, dtype=np.float64)
        self.timesteps = timesteps
        self.box_offsets = frame_offsets + TRR_HEADER_SIZE + 2 * float_size
        self.coordinate_offsets = frame_offsets + TRR_HEADER_SIZE + 2 * float_size + box_size + vir_size + pres_size
        self.nframes = nframes

    cdef PBCBox load_box(self, int frame_id):
        cdef XDRFILE *xfp
        cdef fsl_uint pos = self.box_offsets[frame_id]
        cdef real[:,::1] box = numpy.empty((DIM, DIM), dtype=np.float64)
        cdef float float_box[DIM][DIM]
        cdef real double_box[DIM][DIM]
        cdef int i,j

        xfp = xdrfile_open(self.filename, "rb")
        if xfp == NULL:
            raise IOError("Could not read file: %s" % self.filename)
        if xdrfile_setpos(xfp, pos) == 0:
            raise IOError("Could not set position in file: %s" % self.filename)
        if self.use_double:
            if xdrfile_read_double(double_box[0], DIM*DIM, xfp) != DIM*DIM:
                raise IOError("Could not read matrix")
            for i in range(DIM):
                for j in range(DIM):
                    box[i, j] = double_box[i][j]
        else:
            if xdrfile_read_float(float_box[0], DIM*DIM, xfp) != DIM*DIM:
                raise IOError("Could not read matrix")
            # Convert float to real
            for i in range(DIM):
                for j in range(DIM):
                    box[i, j] = <real> float_box[i][j]

        # Close file and release memory
        xdrfile_close(xfp)

        return PBCBox(box)



    cdef real[:, ::1] load_coords(self, int frame_id, fsl_int[:] atomids) nogil except *:
        cdef XDRFILE *xfp
        cdef fsl_uint pos
        cdef int size = atomids.shape[0]
        cdef int i
        cdef fsl_int atomid
        cdef int natoms, ndecompressed
        cdef real[:, ::1] coords
        cdef rvec real_coords
        cdef fvec float_coords
        cdef int float_size = sizeof(real)
        cdef char *filename

        with gil:
            pos = self.coordinate_offsets[frame_id]
            coords = numpy.empty((size, DIM))
            filename = self.filename

        # Open traj file
        xfp = xdrfile_open(filename, "rb")
        if xfp == NULL:
            with gil:
                raise IOError("Could not read file: %s" % self.filename)

        if self.use_double:
            float_size = sizeof(real)

            # Build the output array
            for i in range(size):
                atomid = atomids[i] - 1 # Reminder: atomids start at 1

                if xdrfile_setpos(xfp, pos + atomid * float_size) == 0:
                    with gil:
                        raise IOError("Could not set position in file: %s" % self.filename)

                # Load coordinates
                if xdrfile_read_double(real_coords, DIM, xfp) < -1:
                    with gil:
                        raise IOError("Could not load coordinates")


                coords[i, XX] = real_coords[XX]
                coords[i, YY] = real_coords[YY]
                coords[i, ZZ] = real_coords[ZZ]

        else:
            float_size = sizeof(float)

            # Build the output array
            for i in range(size):
                atomid = atomids[i] - 1 # Reminder: atomids start at 1

                if xdrfile_setpos(xfp, pos + atomid * float_size * DIM) == 0:
                    with gil:
                        raise IOError("Could not set position in file: %s" % self.filename)

                # Load coordinates
                if xdrfile_read_float(float_coords, DIM, xfp) < -1:
                    with gil:
                        raise IOError("Could not load coordinates")

                coords[i, XX] = float_coords[XX]
                coords[i, YY] = float_coords[YY]
                coords[i, ZZ] = float_coords[ZZ]

        # Close file
        xdrfile_close(xfp)

        return coords

_Coordinates_readers = {".gro": GroReaderCoords,
                        ".xtc": XtcReaderCoords,
                        ".trr": TrrReaderCoords}



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function, absolute_import, division
import os
import sys
import numpy as np
import time

from . import __version__
from .command import Command, AnalyticalCommand, MembraneAnalysisCommand, MembraneProperty
from .util import backup_file


class CmdVersion(Command):
    _name = "version"
    _help = "Displays FATSLiM version"

    def run(self):
        from . import print_full_version_info
        print_full_version_info()


class CmdHelp(Command):
    _name = "help"
    _help = "Displays help about FATSLiM's commands"

    def fill_parser(self):
        super(CmdHelp, self).fill_parser()
        self.parser.add_argument("command", nargs="?",
                                 help="Command to display information about",
                                 default=None)

    def run(self):
        if self.namespace.command is None:
            print(self.registry.usage)

            print("Available commands:")

            cmd_names = list(self.registry.commands.keys())
            cmd_names.sort()

            fmt_str = "  %%-%is%%s" % (max([len(val) for val in cmd_names]) + 4)

            for name in cmd_names:
                print(fmt_str % (name, self.registry.commands[name].desc))

            print("\nSee '%s help <command>' for more information on a specific command" %
                  self.registry.name)

        else:
            try:
                self.registry.commands[self.namespace.command].help()
            except KeyError:
                print("Unkown command: %s. See '%s --help" % (self.namespace.command,
                                                              self.registry.name))


class CmdTest(Command):
    _name = "self-test"
    _help = "Performs FATSLiM's sanity check up"

    def run(self):  # pragma: no cover
        from . import print_full_version_info
        import os

        print_full_version_info(with_header=False)

        try:
            # noinspection PyUnresolvedReferences
            import pytest
        except ImportError:
            self.print_error("pytest is required to run self-test. Please install it")
            return

        try:
            # noinspection PyUnresolvedReferences
            from . import test
        except ImportError:
            self.print_error("Could not find fatslim's test module. Are you sure it is installed?")
            return

        ret_code = pytest.main(["-rxs", "--tb=short", "%s" % os.path.dirname(test.__file__)])
        if ret_code == 0:
            print("\nOK buddy, I am ready to rock!")
        else:
            print("Ouch... Congratulations! you probably found a bug!")
            print("Please contact the FATSLiM devs with this output so they can work it out!")


class CmdAggregates(AnalyticalCommand):
    _name = "aggregates"
    _help = "Identifies and reports aggregates"

    def __init__(self, registry):
        super(CmdAggregates, self).__init__(registry)
        self.num_aggregates = None

    def fill_parser(self):
        super(CmdAggregates, self).fill_parser()

        self.parser_output_group.add_argument("--output", "-o", default=None, type=str,
                                              help="Plot number of aggregates over the selected "
                                                   "trajectory frames")

        self.parser_output_group.add_argument("--output-index-hg", default=None, type=str,
                                              help="Index group to store aggregates headgroups")

        self.parser_output_group.add_argument("--output-index", default=None, type=str,
                                              help="Index group to store aggregates")

        self.parser_analysis_group.add_argument("--cutoff", default=2.0, type=float,
                                                help="Cutoff distance for aggregate identication")

    def prepare_results(self, num_frames):
        self.num_aggregates = np.zeros(num_frames, dtype=int)
        self.timesteps = np.zeros(num_frames, dtype=int)

    def process_frame(self, frame):
        aggregates = frame.get_aggregates(self.namespace.cutoff)

        self.timesteps[self.processing_index] = frame.timestep

        self.num_aggregates[self.processing_index] = len(aggregates)

        n_aggregates = 0
        if self.namespace.output_index is not None:
            fname, ext = os.path.splitext(self.namespace.output_index)
            fname = "%s_%04i.ndx" % (fname, frame.timestep * .001)

            with open(fname, "w") as fp:
                for agg in aggregates:
                    n_aggregates += 1
                    fp.write("[ aggregate_%i ]\n" % n_aggregates)

                    slice_i = 0
                    atomids = agg.lipid_atomids[slice_i:slice_i + 15]

                    while len(atomids) > 0:
                        fp.write(" ".join(["%i" % val for val in atomids]))
                        fp.write("\n")
                        slice_i += 15
                        atomids = agg.lipid_atomids[slice_i:slice_i + 15]

                    fp.write("\n")

        n_aggregates = 0
        if self.namespace.output_index_hg is not None:
            fname, ext = os.path.splitext(self.namespace.output_index_hg)
            fname = "%s_%04i.ndx" % (fname, frame.timestep * .001)

            with open(fname, "w") as fp:
                for agg in aggregates:
                    n_aggregates += 1
                    fp.write("[ aggregate_%i ]\n" % n_aggregates)

                    slice_i = 0
                    atomids = agg.hg_atomids[slice_i:slice_i + 15]

                    while len(atomids) > 0:
                        fp.write(" ".join(["%i" % val for val in atomids]))
                        fp.write("\n")
                        slice_i += 15
                        atomids = agg.hg_atomids[slice_i:slice_i + 15]

                    fp.write("\n")

    def process_results(self):
        filepath = self.namespace.output
        if filepath is not None:
            fname, ext = os.path.splitext(filepath)

            if ext != ".xvg":
                filepath = "%s.xvg" % fname

            backup_file(filepath)

            with open(filepath, "w") as fp:
                fp.write(self.get_xvg_header())
                fp.write("@ title \"Number of aggregates\"\n")
                fp.write("@ xaxis label \"Time (ps)\"\n")
                fp.write("@ yaxis label \"Number of aggregates\"\n")
                fp.write("@TYPE xy\n")

                for i, val in enumerate(self.timesteps):
                    fp.write("% 8.3f    % 4i\n" % (val, self.num_aggregates[i]))


class CmdMembranes(AnalyticalCommand):
    _name = "membranes"
    _help = "Identifies and report membranes"

    def __init__(self, registry):
        super(CmdMembranes, self).__init__(registry)
        self.num_membranes = None

    def fill_parser(self):
        super(CmdMembranes, self).fill_parser()

        self.parser_output_group.add_argument("-o", "--output", default=None, type=str,
                                              help="Plot number of membranes over the "
                                                   "selected trajectory frames")

        self.parser_output_group.add_argument("--output-index-hg", default=None, type=str,
                                              help="Index group to store leaflet headgroups")

        self.parser_output_group.add_argument("--output-index", default=None, type=str,
                                              help="Index group to store leaflets")

        self.parser_analysis_group.add_argument("--cutoff", default=2.0, type=float,
                                                help="Cutoff distance (in nm) for leaflet "
                                                     "identification")

    def prepare_results(self, num_frames):
        self.num_membranes = np.zeros(num_frames, dtype=int)
        self.timesteps = np.zeros(num_frames, dtype=int)

    def process_frame(self, frame):
        output = super(CmdMembranes, self).process_frame(frame)

        try:
            membranes = frame.get_membranes(self.namespace.cutoff, update=True)
        except (ValueError, IndexError):  # pragma: no cover
            output += "->No membrane found!"
        else:
            n_membranes = len(membranes)
            output = "->%i membrane%s found!\n" % (n_membranes, "s"[n_membranes == 1:])

        self.num_membranes[self.processing_index] = n_membranes

        if self.namespace.output_index is not None:
            fname, ext = os.path.splitext(self.namespace.output_index)
            fname = "%s_%04i.ndx" % (fname, frame.timestep * .001)

            with open(fname, "w") as fp:
                for mid, membrane in enumerate(membranes):
                    for i, leaf in enumerate(membrane):
                        fp.write("[ membrane_%i_leaflet_%i ]\n" % (mid + 1, i + 1))

                        slice_i = 0
                        atomids = leaf.lipid_atomids[slice_i:slice_i + 15]

                        while len(atomids) > 0:
                            fp.write(" ".join(["%i" % val for val in atomids]))
                            fp.write("\n")
                            slice_i += 15
                            atomids = leaf.lipid_atomids[slice_i:slice_i + 15]

                        fp.write("\n")
            output += "leaflets saved to '%s'\n" % fname

        if self.namespace.output_index_hg is not None:
            fname, ext = os.path.splitext(self.namespace.output_index_hg)
            fname = "%s_%04i.ndx" % (fname, frame.timestep * .001)

            with open(fname, "w") as fp:
                for mid, membrane in enumerate(membranes):
                    for i, leaf in enumerate(membrane):
                        fp.write("[ membrane_%i_leaflet_%i ]\n" % (mid + 1, i + 1))

                        slice_i = 0
                        atomids = leaf.hg_atomids[slice_i:slice_i + 15]

                        while len(atomids) > 0:
                            fp.write(" ".join(["%i" % val for val in atomids]))
                            fp.write("\n")
                            slice_i += 15
                            atomids = leaf.hg_atomids[slice_i:slice_i + 15]

                        fp.write("\n")
            output += "headgroup atoms saved to '%s'\n" % fname
        return output

    def process_results(self):
        filepath = self.namespace.output
        if filepath is not None:
            fname, ext = os.path.splitext(filepath)

            filepath = "%s.xvg" % fname

            backup_file(filepath)

            with open(filepath, "w") as fp:
                fp.write(self.get_xvg_header())
                fp.write("@ title \"Number of membranes\"\n")
                fp.write("@ xaxis label \"Time (ps)\"\n")
                fp.write("@ yaxis label \"Number of membranes\"\n")
                fp.write("@TYPE xy\n")

                for i, val in enumerate(self.timesteps):
                    fp.write("% 8.3f    % 4i\n" % (val, self.num_membranes[i]))


class PropertyThickness(MembraneProperty):
    _fullname = "thickness"
    _shortname = "thickness"
    _units = "nm"

    def fill_parsers(self, parser_analysis, parser_input, parser_output):
        super(PropertyThickness, self).fill_parsers(parser_analysis, parser_input, parser_output)

        parser_analysis.add_argument("--thickness-cutoff",
                                     dest="thickness_cutoff", default=6.0, type=float,
                                     help="Cutoff distance (in nm) used to identify inter-leaflet "
                                          "neighbors")

    def get_values(self, membrane):
        mem_val, l1_val, l2_val = membrane.get_thickness(interleaflet_cutoff=
                                                         self.parent.namespace.thickness_cutoff,
                                                         only_average=True,
                                                         force=False)

        self.membrane_avg_values[self.parent.processing_index] = mem_val
        self.leaflet1_avg_values[self.parent.processing_index] = l1_val[0]
        self.leaflet2_avg_values[self.parent.processing_index] = l2_val[0]

        return mem_val, l1_val[0], l2_val[0]

    def get_raw_values(self, membrane):
        mem_val, l1_val, l2_val = membrane.get_thickness(interleaflet_cutoff=
                                                         self.parent.namespace.thickness_cutoff,
                                                         only_average=False,
                                                         force=False)
        return l1_val, l2_val


class CmdThickness(MembraneAnalysisCommand):
    _name = "thickness"
    _help = "Retrieves bilayer thickness"
    _property_classes = [PropertyThickness]


class PropertyAPL(MembraneProperty):
    _fullname = "area per lipid"
    _shortname = "apl"
    _units = "nm^2"
    _groupable_by_type = True

    def fill_parsers(self, parser_analysis, parser_input, parser_output):
        super(PropertyAPL, self).fill_parsers(parser_analysis, parser_input, parser_output)

        parser_analysis.add_argument("--apl-cutoff",
                                     dest="apl_cutoff", default=3.0, type=float,
                                     help="Cutoff distance (in nm) used to approximate "
                                          "planar region")

        parser_analysis.add_argument("--apl-limit",
                                     dest="apl_limit", default=10.0, type=float,
                                     help="Upper limit (in nm2) considered when calculating "
                                          "individual APL values")

    def get_values(self, membrane):
        mem_val, l1_val_dict, l2_val_dict = membrane.get_apl(
            cutoff=self.parent.namespace.apl_cutoff,
            area_limit=self.parent.namespace.apl_limit,
            by_type=self.needs_grouping,
            only_average=True,
            force=False)

        if self.needs_grouping:
            mem_val_by_type = {}
            mem_val = [0.0, 0]
            l1_val = [0.0, 0]
            l2_val = [0.0, 0]

            for key, val in l1_val_dict.items():
                mem_val[0] += val[0] * val[1]
                mem_val[1] += val[0]

                l1_val[0] += val[0] * val[1]
                l1_val[1] += val[0]

                try:
                    self.leaflet1_avg_values_by_type[key][self.parent.processing_index] = val[1]
                except KeyError:
                    self.leaflet1_avg_values_by_type[key] = np.zeros(len(self.membrane_avg_values))
                    self.leaflet1_avg_values_by_type[key][self.parent.processing_index] = val[1]

                mem_val_by_type[key] = [val[1] * val[0], val[0]]

            for key, val in l2_val_dict.items():
                mem_val[0] += val[0] * val[1]
                mem_val[1] += val[0]

                l2_val[0] += val[0] * val[1]
                l2_val[1] += val[0]

                try:
                    self.leaflet2_avg_values_by_type[key][self.parent.processing_index] = val[1]
                except KeyError:
                    self.leaflet2_avg_values_by_type[key] = np.zeros(len(self.membrane_avg_values))
                    self.leaflet2_avg_values_by_type[key][self.parent.processing_index] = val[1]

                try:
                    mem_val_by_type[key][0] += val[1] * val[0]
                    mem_val_by_type[key][1] += val[0]
                except KeyError:  # pragma: no cover
                    mem_val_by_type[key] = [val[1] * val[0], val[0]]

            for key, val in mem_val_by_type.items():
                val = float(val[0]) / val[1]

                try:
                    self.membrane_avg_values_by_type[key][self.parent.processing_index] = val
                except KeyError:
                    self.membrane_avg_values_by_type[key] = np.zeros(len(self.membrane_avg_values))
                    self.membrane_avg_values_by_type[key][self.parent.processing_index] = val

            mem_val = float(mem_val[0]) / mem_val[1]
            l1_val = float(l1_val[0]) / l1_val[1]
            l2_val = float(l2_val[0]) / l2_val[1]

            self.membrane_avg_values[self.parent.processing_index] = mem_val
            self.leaflet1_avg_values[self.parent.processing_index] = l1_val
            self.leaflet2_avg_values[self.parent.processing_index] = l2_val

        else:
            l1_val = l1_val_dict[0]
            l2_val = l2_val_dict[0]

            self.membrane_avg_values[self.parent.processing_index] = mem_val
            self.leaflet1_avg_values[self.parent.processing_index] = l1_val
            self.leaflet2_avg_values[self.parent.processing_index] = l2_val

        return mem_val, l1_val, l2_val

    def get_raw_values(self, membrane):
        mem_val, l1_val, l2_val = membrane.get_apl(cutoff=self.parent.namespace.apl_cutoff,
                                                   area_limit=self.parent.namespace.apl_limit,
                                                   by_type=False,
                                                   only_average=False,
                                                   force=False)

        return l1_val, l2_val


class PropertyArea(MembraneProperty):
    _fullname = "area"
    _shortname = "area"
    _units = "nm^2"
    _groupable_by_type = False
    _raw_values_supported = False

    def get_values(self, membrane):
        mem_val, l1_val, l2_val = membrane.get_apl(
            cutoff=self.parent.namespace.apl_cutoff,
            area_limit=self.parent.namespace.apl_limit,
            by_type=self.needs_grouping,
            only_average=True,
            force=False)

        l1_val = l1_val[3]
        l2_val = l2_val[3]

        self.membrane_avg_values[self.parent.processing_index] = 0.5 * (l1_val + l2_val)
        self.leaflet1_avg_values[self.parent.processing_index] = l1_val
        self.leaflet2_avg_values[self.parent.processing_index] = l2_val

        return mem_val, l1_val, l2_val

    def get_raw_values(self, membrane):  # pragma: no cover
        pass


class CmdApl(MembraneAnalysisCommand):
    _name = "apl"
    _help = "Retrieves area per lipid"
    _property_classes = [PropertyAPL, PropertyArea]


class CmdBenchmark(AnalyticalCommand):
    _name = "benchmark"
    _help = "Benchmark internal routine (not intented to be a end user command!)"

    def run(self):  # pragma: no cover
        import time
        from .datareading import load_trajectory
        from .core_base import pretty_delta, pretty_duration
        import gc

        def get_memory_usage():
            memory = -1
            with open('/proc/self/status', "rb") as fp:
                for line in fp:
                    if line.startswith("VmRSS:"):
                        memory = int(line.split()[1])
                        break
            return memory

        def h_memory(memory):
            units = ["kB", "MB", "GB", "TB"]
            unitid = 0
            while memory > 1024:
                memory /= 1024.
                unitid += 1
            return "%.1f %s" % (memory, units[unitid])

        mem_before = get_memory_usage()
        time_before = time.time()
        print("Memory usage BEFORE starting: %s" % h_memory(mem_before))

        print("\nInitializing Trajectory...")
        mem_begin = get_memory_usage()
        time_begin = time.time()
        traj = load_trajectory(self.namespace.conf,
                               self.namespace.index,
                               self.namespace.trajectory,
                               self.verbose)
        time_end = time.time()
        mem_end = get_memory_usage()
        print("=> Memory usage after initialization: %s (increment: %s) - Time taken: %s" %
              (h_memory(mem_end),
               h_memory(mem_end - mem_begin),
               pretty_delta(time_begin, time_end)))

        print("\nPreloading Trajectory...")
        mem_begin = get_memory_usage()
        time_begin = time.time()
        traj.initialize(hg_group=self.namespace.hg_group)
        time_end = time.time()
        mem_end = get_memory_usage()
        print("=> Memory usage after preloading: %s (increment: %s) - Time taken: %s" %
              (h_memory(mem_end),
               h_memory(mem_end - mem_begin),
               pretty_delta(time_begin, time_end)))

        size = len(traj)
        first_frame = max(0, self.namespace.begin)
        last_frame = min(size, self.namespace.end)
        if last_frame < 0:
            last_frame = size
        num_frames = last_frame - first_frame

        if num_frames <= 0:
            return

        getting_times = np.zeros(num_frames)
        getting_mems = np.zeros(num_frames)
        coords_times = np.zeros(num_frames)
        coords_mems = np.zeros(num_frames)
        coords_bbox_times = np.zeros(num_frames)
        coords_bbox_mems = np.zeros(num_frames)
        directions_times = np.zeros(num_frames)
        directions_mems = np.zeros(num_frames)
        normals_times = np.zeros(num_frames)
        normals_mems = np.zeros(num_frames)
        membrane_times = np.zeros(num_frames)
        membrane_mems = np.zeros(num_frames)
        apl_times = np.zeros(num_frames)
        apl_mems = np.zeros(num_frames)
        thickness_times = np.zeros(num_frames)
        thickness_mems = np.zeros(num_frames)
        gc_times = np.zeros(num_frames)
        gc_mems = np.zeros(num_frames)
        total_frame_times = np.zeros(num_frames)

        full_details = False
        for i in range(first_frame, last_frame):
            if full_details:
                print("\nProcessing frame #%i (current memory usage: %s):" %
                      (
                          i - first_frame + 1,
                          h_memory(get_memory_usage())
                      ))
            else:
                print("Processing frame #%i (current memory usage: %s)" %
                      (
                          i - first_frame + 1,
                          h_memory(get_memory_usage())
                      ), end="")
            mem_begin = get_memory_usage()
            time_begin = time.time()
            frame = traj[i]
            time_end = time.time()
            mem_end = get_memory_usage()
            frame_begin = time_begin
            if full_details:
                print("  => Memory usage after getting frame: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end)))
            getting_times[i - first_frame] = time_end - time_begin
            getting_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting coords
            mem_begin = get_memory_usage()
            time_begin = time.time()
            frame.bead_coords
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting coords: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end),))
            coords_times[i - first_frame] = time_end - time_begin
            coords_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting coords
            mem_begin = get_memory_usage()
            time_begin = time.time()
            frame.bead_coords_bbox
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting coords in bbox : %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end),))
            coords_bbox_times[i - first_frame] = time_end - time_begin
            coords_bbox_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting directions
            mem_begin = get_memory_usage()
            time_begin = time.time()
            frame.directions
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting directions: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end),
                       ))
            directions_times[i - first_frame] = time_end - time_begin
            directions_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting normals
            mem_begin = get_memory_usage()
            time_begin = time.time()
            frame.normals
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting normals: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end)
                       ))
            normals_times[i - first_frame] = time_end - time_begin
            normals_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting membranes
            mem_begin = get_memory_usage()
            time_begin = time.time()
            membrane = frame.get_membranes()[0]
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting membranes: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end)))
            membrane_times[i - first_frame] = time_end - time_begin
            membrane_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting APL
            mem_begin = get_memory_usage()
            time_begin = time.time()
            membrane.get_apl()
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting APL: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end)))
            apl_times[i - first_frame] = time_end - time_begin
            apl_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Getting Thickness
            mem_begin = get_memory_usage()
            time_begin = time.time()
            membrane.get_thickness()
            time_end = time.time()
            mem_end = get_memory_usage()
            if full_details:
                print("  => Memory usage after getting thickness: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end)))
            thickness_times[i - first_frame] = time_end - time_begin
            thickness_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

            # Garbage collection
            mem_begin = get_memory_usage()
            time_begin = time.time()
            try:
                del frame
                del membrane
            except UnboundLocalError:
                pass
            gc.collect()
            time_end = time.time()
            mem_end = get_memory_usage()
            frame_end = time_end
            if full_details:
                print("  => Memory usage after garbage collection: %s (increment: %s) - "
                      "Time taken: %s" %
                      (h_memory(mem_end),
                       h_memory(mem_end - mem_begin),
                       pretty_delta(time_begin, time_end)))
            else:
                print(" Done in %s" % pretty_delta(frame_begin, frame_end))
            gc_times[i - first_frame] = time_end - time_begin
            gc_mems[i - first_frame] = mem_end - mem_begin
            total_frame_times[i - first_frame] += time_end - time_begin

        mem_after = get_memory_usage()
        time_after = time.time()
        print("\nGrand total summary: Memory usage: %s (increment: %s) - Time taken: %s" %
              (h_memory(mem_after),
               h_memory(mem_after - mem_before),
               pretty_delta(time_before, time_after))
              )
        print("Average values over %i frames:" % num_frames)
        print("Average processing time per frame: %s-+%s" %
              (pretty_duration(total_frame_times.mean()),
               pretty_duration(total_frame_times.std())))
        print("Getting frames: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(getting_mems.mean()),
                  pretty_duration(getting_times.mean()),
                  (getting_times / total_frame_times * 100.0).mean()
              ))
        print("Getting coords: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(coords_mems.mean()),
                  pretty_duration(coords_times.mean()),
                  (coords_times / total_frame_times * 100.0).mean()
              ))
        print("Getting coords_bbox: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(coords_bbox_mems.mean()),
                  pretty_duration(coords_bbox_times.mean()),
                  (coords_bbox_times / total_frame_times * 100.0).mean()
              ))
        print("Total I/O: %.1f%% of frame processing" %
              ((getting_times + coords_times + coords_bbox_times) /
               total_frame_times * 100.0).mean())
        print("Getting directions: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(directions_mems.mean()),
                  pretty_duration(directions_times.mean()),
                  (directions_times / total_frame_times * 100.0).mean()
              ))
        print("Getting normals: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(normals_mems.mean()),
                  pretty_duration(normals_times.mean()),
                  (normals_times / total_frame_times * 100.0).mean()
              ))
        print("Getting membrane: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(membrane_mems.mean()),
                  pretty_duration(membrane_times.mean()),
                  (membrane_times / total_frame_times * 100.0).mean()
              ))
        print("Getting apl: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(apl_mems.mean()),
                  pretty_duration(apl_times.mean()),
                  (apl_times / total_frame_times * 100.0).mean()
              ))
        print("Getting thickness: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(thickness_mems.mean()),
                  pretty_duration(thickness_times.mean()),
                  (thickness_times / total_frame_times * 100.0).mean()
              ))
        print("Garbage collection: Memory usage: %s - Time taken: %s (%.1f%% frame processing)" %
              (
                  h_memory(gc_mems.mean()),
                  pretty_duration(gc_times.mean()),
                  (gc_times / total_frame_times * 100.0).mean()
              ))



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
#cython: cdivision=True
#cython: boundscheck=False
#cython: profile=False

from libc.stdlib cimport malloc, realloc, free, abort
from libc.math cimport sqrt, fabs
#cimport numpy as np
import numpy as np

DEF POLYGON_ALLOCATION_INCREMENT = 10
DEF ZOI_DEFAULT_SIZE = 1e6
DEF XX = 0
DEF YY = 1
DEF EPSILON = 1e-6

# Useful functions
cdef bint fast_inside_clipped_polygon(real_point p, real_point centerclip, real_point cp1, real_point cp2) nogil:
    cdef real winding_ok, winding_test
    winding_ok = (cp2[XX]-cp1[XX])*(centerclip[YY]-cp1[YY]) - (cp2[YY]-cp1[YY])*(centerclip[XX]-cp1[XX])
    winding_test = (cp2[XX]-cp1[XX])*(p[YY]-cp1[YY]) - (cp2[YY]-cp1[YY])*(p[XX]-cp1[XX])

    if fabs(winding_test) < EPSILON:
        return False

    return winding_ok / winding_test > 0


cdef void fast_compute_intersection(real_point s, real_point e, real_point cp1, real_point cp2, real_point intersection) nogil:
    cdef real n1, n2, n3
    cdef real_point dc, dp
    dc[XX] = cp1[0] - cp2[0]
    dc[YY] = cp1[1] - cp2[1]
    dp[XX] = s[0] - e[0]
    dp[YY] = s[1] - e[1]

    n1 = cp1[0] * cp2[1] - cp1[1] * cp2[0]
    n2 = s[0] * e[1] - s[1] * e[0]
    n3 = (dc[0] * dp[1] - dc[1] * dp[0])

    if fabs(n3) < EPSILON:
        intersection[XX] = s[XX]
        intersection[YY] = s[YY]
    else:
        n3 = 1.0 / n3
        intersection[XX] = (n1 * dp[0] - n2 * dc[0]) * n3
        intersection[YY] = (n1 * dp[1] - n2 * dc[1]) * n3


# See: http://geomalgorithms.com/a01-_area.html
cdef inline int point_is_left(real_point l0, real_point l1, real_point p) nogil:
    cdef real is_left = (l1[XX] - l0[XX]) * (p[YY] - l0[YY]) - (p[XX] -  l0[XX]) * (l1[YY] - l0[YY])

    if fabs(is_left) < EPSILON:
        return 0 # Point p is on the line

    if is_left > 0:
        return 1 # Point p is on the left

    return -1 # Point p is on the right
########################################################################################################################
# Polygon pseudo class methods
########################################################################################################################
cdef Polygon *polygon_new() nogil:
    cdef Polygon *self = NULL
    self = <Polygon *> malloc(sizeof(Polygon))
    if self == NULL:
        abort()
    self.size = 0
    self.allocated_size = 0
    self.points = NULL
    return self

cdef Polygon *polygon_new_from_memview(real[:, ::1] coords) nogil:
    cdef Polygon *self = polygon_new()
    cdef int size = coords.shape[0]
    cdef int i

    self.points = <real_point *> malloc(size * sizeof(real_point))
    if self.points == NULL:
        abort()

    self.allocated_size = size

    for i in range(size):
        self.points[i][XX] = coords[i, XX]
        self.points[i][YY] = coords[i, YY]
    self.size = size

    return self

cdef Polygon *polygon_new_from_polygon( Polygon *other) nogil:
    cdef Polygon *self = polygon_new()

    polygon_copy(self, other)

    return self

cdef void polygon_copy(Polygon *self, Polygon *other) nogil:
    cdef int size = other.size
    cdef int i

    if self.allocated_size < size:
        if self.allocated_size == 0:
            self.points = <real_point *> malloc(size * sizeof(real_point))
        else:
            self.points = <real_point *> realloc(<void **> self.points, size * sizeof(real_point))

        self.allocated_size = size

    if self.points == NULL:
        abort()

    for i in range(size):
        self.points[i][XX] = other.points[i][XX]
        self.points[i][YY] = other.points[i][YY]
    self.size = size

cdef void polygon_get_cog(Polygon *self, real_point cog) nogil:
    cdef int i

    cog[XX] = 0
    cog[YY] = 0

    for i in range(self.size):
        cog[XX] += self.points[i][XX]
        cog[YY] += self.points[i][YY]

    if self.size == 0:
        return

    cog[XX] /= self.size
    cog[YY] /= self.size


cdef void polygon_destroy(Polygon *self) nogil:
    if self == NULL:
        return
    free(self.points)
    free(self)


cdef void polygon_empty(Polygon *self) nogil:
    self.size = 0


cdef void polygon_append(Polygon *self, real_point point) nogil:
    if self.size + 1 > self.allocated_size:
        self.allocated_size += POLYGON_ALLOCATION_INCREMENT
        if self.points == NULL:
            self.points = <real_point *> malloc(sizeof(real_point) * self.allocated_size)
        else:
            self.points = <real_point *> realloc(<void **> self.points, sizeof(real_point) * self.allocated_size)

        if self.points == NULL:
            abort()

    self.points[self.size][XX] = point[XX]
    self.points[self.size][YY] = point[YY]

    self.size += 1


cdef real polygon_get_area(Polygon *self) nogil:
    cdef int i, previous_i
    cdef real area = 0

    if self == NULL:
        return -1.0

    if self.size < 3:
        return -1.0

    previous_i = self.size - 1
    for i in range(0, self.size):
        area += self.points[previous_i][XX] * self.points[i][YY] - self.points[i][XX] * self.points[previous_i][YY]
        previous_i = i

    if area < 0:
        return area * -0.5
    else:
        return area * 0.5

cdef real polygon_get_perimeter(Polygon *self) nogil:
    cdef int i, previous_i
    cdef real perimeter = 0, delta_x, delta_y

    if self.size < 3:
        return 0.0

    previous_i = self.size - 1
    for i in range(0, self.size):
        delta_x = self.points[i][XX] - self.points[previous_i][XX]
        delta_y = self.points[i][YY] - self.points[previous_i][YY]
        perimeter += sqrt(delta_x * delta_x + delta_y * delta_y)
        previous_i = i

    return perimeter


cdef object polygon_as_array(Polygon *self):
    cdef int i
    cdef real[:, ::1] py_self = np.empty([self.size, 2])

    for i in range(self.size):
        py_self[i, XX] = self.points[i][XX]
        py_self[i, YY] = self.points[i][YY]

    return np.asarray(py_self)



cdef Polygon *fast_get_clipped_polygon(Polygon *subject, Polygon *clipper) nogil:
    cdef real_point clipper_cog, cp1, cp2, s, e, intersect
    cdef int i, j
    cdef Polygon *clipped = NULL
    cdef Polygon *toclip = polygon_new()

    if subject.size < 3 or clipper.size < 3:
        return toclip

    clipped = polygon_new_from_polygon(subject)

    # Compute the center of geometry of the clipping polygon
    polygon_get_cog(clipper, clipper_cog)

    cp1[XX] = clipper.points[clipper.size -1][XX]
    cp1[YY] = clipper.points[clipper.size -1][YY]
    for i in range(clipper.size):
        cp2[XX] = clipper.points[i][XX]
        cp2[YY] = clipper.points[i][YY]

        # Copy clipped polygon to polygon to be clipped
        polygon_copy(toclip, clipped)

        # Check if we need to break
        if toclip.size == 0:
            break

        s[XX] = toclip.points[toclip.size - 1][XX]
        s[YY] = toclip.points[toclip.size - 1][YY]

        # Empty clipped polygon
        polygon_empty(clipped)

        for j in range(toclip.size):
            e[XX] = toclip.points[j][XX]
            e[YY] = toclip.points[j][YY]

            if fast_inside_clipped_polygon(e, clipper_cog, cp1, cp2):
                if not fast_inside_clipped_polygon(s, clipper_cog, cp1, cp2):
                    # Append intersection to clipped polygon
                    fast_compute_intersection(s, e, cp1, cp2, intersect)
                    polygon_append(clipped, intersect)
                # Append current point to clipped polygon
                polygon_append(clipped, e)
            else:
                if fast_inside_clipped_polygon(s, clipper_cog, cp1, cp2):
                    # Append intersection to clipped polygon
                    fast_compute_intersection(s, e, cp1, cp2, intersect)
                    polygon_append(clipped, intersect)

            s[XX] = e[XX]
            s[YY] = e[YY]

        cp1[XX] = cp2[XX]
        cp1[YY] = cp2[YY]

    # Free memory
    polygon_destroy(toclip)

    return clipped

# See: http://geomalgorithms.com/a03-_inclusion.html#wn_PnPoly%28%29
cdef bint polygon_is_inside(Polygon *self, real_point p) nogil:
    cdef int i
    cdef int wn
    cdef real_point next_vertex, current_vertex

    current_vertex[XX] = self.points[self.size - 1][XX] # Last vertex
    current_vertex[YY] = self.points[self.size - 1][YY] # Last vertex

    wn = 0
    for i in range(self.size):
        next_vertex[XX] = self.points[i][XX]
        next_vertex[YY] = self.points[i][YY]

        if current_vertex[YY] <= p[YY]:
            if next_vertex[YY] > p[YY]: # upward crossing
                if point_is_left(current_vertex, next_vertex, p) > 0: # p is on the left side of the edge
                    wn += 1
        else:
            if next_vertex[YY] <= p[YY]: # downward crossing
                if point_is_left(current_vertex, next_vertex, p) < 0: # p is on the right side of the edge
                    wn -= 1

        current_vertex[XX] = next_vertex[XX]
        current_vertex[YY] = next_vertex[YY]

    return wn != 0


########################################################################################################################
#
# Python API
#
########################################################################################################################

def get_polygon_area(real[:,::1] points not None):
    cdef Polygon *polygon = NULL
    cdef real area

    with nogil:
        polygon = polygon_new_from_memview(points)
        area = polygon_get_area(polygon)

        # Free memory
        polygon_destroy(polygon)

    return area

def get_polygon_perimeter(real[:,::1] points not None):
    cdef Polygon *polygon = NULL
    cdef real perimeter

    with nogil:
        polygon = polygon_new_from_memview(points)
        perimeter = polygon_get_perimeter(polygon)

        # Free memory
        polygon_destroy(polygon)

    return perimeter

def get_clipped_polygon(real[:, ::1] subject_points not None, real[:, ::1] clipper_points not None):
    cdef Polygon *subject = polygon_new_from_memview(subject_points)
    cdef Polygon *clipper = polygon_new_from_memview(clipper_points)
    cdef Polygon *clipped = fast_get_clipped_polygon(subject, clipper)

    clipped_py = polygon_as_array(clipped)

    # Free memory
    polygon_destroy(clipper)
    polygon_destroy(subject)
    polygon_destroy(clipped)

    return clipped_py


def is_inside_polygon(real[:, ::1] points, real[:] point):
    cdef Polygon *polygon = polygon_new_from_memview(points)
    cdef real_point p
    cdef bint is_inside
    p[XX] = point[XX]
    p[YY] = point[YY]

    is_inside = polygon_is_inside(polygon, p)

    polygon_destroy(polygon)

    return is_inside

cdef bint same_side(real_point a, real_point b, real_point x_line, real_point y_line) nogil:
    return ((x_line[YY] - y_line[YY])*(a[XX]-x_line[XX]) + (y_line[XX] - x_line[XX])*(a[YY] - x_line[YY])) * \
    ((x_line[YY] - y_line[YY])*(b[XX]-x_line[XX]) + (y_line[XX] - x_line[XX])*(b[YY] - x_line[YY])) >= 0

cdef bint are_parallel(real_point a1, real_point a2, real_point b1, real_point b2) nogil:
    cdef real det = (a1[XX] - a2[XX]) * (b1[YY] - b2[YY]) - (a1[YY] - a2[YY])* (b1[XX] - b2[XX])
    return fabs(det) < EPSILON

cdef void get_intersection(real_point a1, real_point a2, real_point b1, real_point b2,
                           real_point *inter_sect) nogil:
    cdef real det = (a1[XX] - a2[XX]) * (b1[YY] - b2[YY]) - (a1[YY] - a2[YY])* (b1[XX] - b2[XX])
    cdef real val1, val2, x, y

    if fabs(det) < EPSILON: # Lines are coincident
        inter_sect[0][XX] = a1[XX]
        inter_sect[0][YY] = a1[YY]
    else:
        val1 = (a1[0] * a2[1] - a1[1] * a2[0])
        val2 = (b1[0] * b2[1] - b1[1] * b2[0])

        x = val1 * (b1[0] - b2[0]) - (a1[0] - a2[0]) * val2
        x /= det

        y = val1 * (b1[1] - b2[1]) - (a1[1] - a2[1]) * val2
        y /= det

        inter_sect[0][XX] = x
        inter_sect[0][YY] = y

cdef void fast_clip_zoi(Polygon *zoi, real_point ref_pt, real_point clipping_pt, Polygon *buffer) nogil:
    cdef real_point middle_pt, other_pt, delta, line_dir, inter_pt
    cdef int first_vid, second_vid
    cdef bint second_same_side
    cdef bint need_cleanup = False

    if buffer == NULL:
        need_cleanup = True
        buffer = polygon_new()

    # Calculate stuff related to clipping line
    middle_pt[XX] = 0.5 * (clipping_pt[XX] + ref_pt[XX])
    middle_pt[YY] = 0.5 * (clipping_pt[YY] + ref_pt[YY])

    delta[XX] = ref_pt[XX] - clipping_pt[XX]
    delta[YY] = ref_pt[YY] - clipping_pt[YY]

    line_dir[XX] = delta[YY]
    line_dir[YY] = -delta[XX]

    other_pt[XX] = middle_pt[XX] + line_dir[XX]
    other_pt[YY] = middle_pt[YY] + line_dir[YY]

    polygon_empty(buffer)

    # Check if line intersects with the polygon
    for second_vid in range(zoi.size):
        if second_vid == 0:
            first_vid = zoi.size - 1
        else:
            first_vid = second_vid - 1


        second_same_side = same_side(ref_pt, zoi.points[second_vid], middle_pt, other_pt)

        if same_side(zoi.points[first_vid], zoi.points[second_vid], middle_pt, other_pt):
            if not second_same_side:
                continue
        else:

            if not are_parallel(zoi.points[first_vid], zoi.points[second_vid], middle_pt, other_pt):
                # Get the intersection point
                get_intersection(zoi.points[first_vid], zoi.points[second_vid], middle_pt, other_pt, &inter_pt)

                polygon_append(buffer, inter_pt)

        if second_same_side:
            polygon_append(buffer, zoi.points[second_vid])


    # Update zoi
    polygon_copy(zoi, buffer)


    # Destroy buffer polygon if needed
    if need_cleanup:
        polygon_destroy(buffer)


cdef Polygon *fast_get_zoi(real_point ref_pt, real_point *pts, int size) nogil:
    cdef Polygon *zoi = polygon_new()
    cdef Polygon *buffer = polygon_new()
    cdef int i
    cdef real_point tmp_pt

    # Start with a big square
    tmp_pt[XX] = -ZOI_DEFAULT_SIZE
    tmp_pt[YY] = -ZOI_DEFAULT_SIZE
    polygon_append(zoi, tmp_pt)
    tmp_pt[XX] = ZOI_DEFAULT_SIZE
    polygon_append(zoi, tmp_pt)
    tmp_pt[YY] = ZOI_DEFAULT_SIZE
    polygon_append(zoi, tmp_pt)
    tmp_pt[XX] = -ZOI_DEFAULT_SIZE
    polygon_append(zoi, tmp_pt)

    # Clip the ZoI using all the points
    for i in range(size):
        fast_clip_zoi(zoi, ref_pt, pts[i], buffer)

    # Release memory
    polygon_destroy(buffer)
    return zoi



###################################################################################################
#
# Python API
#
###################################################################################################

def get_zoi(real[:] ref_pt, real[:,::1] pts):
    cdef real_point ref_pt_c
    cdef real_point *pts_c = NULL
    cdef int size = pts.shape[0]
    cdef int i
    cdef Polygon *zoi = NULL

    with nogil:
        # Convert Python to C
        ref_pt_c[XX] = ref_pt[XX]
        ref_pt_c[YY] = ref_pt[YY]

        pts_c = <real_point *> malloc(size * sizeof(real_point))
        if pts_c == NULL:
            abort()

        for i in range(size):
            pts_c[i][XX] = pts[i, XX]
            pts_c[i][YY] = pts[i, YY]

        # Get ZoI
        zoi = fast_get_zoi(ref_pt_c, pts_c, size)

        # Free memory
        free(pts_c)

    zoi_py = polygon_as_array(zoi)

    polygon_destroy(zoi)

    return zoi_py

def clip_zoi(real[:, ::1] zoi, real[:] ref_pt, real[:] clipping_pt):
    cdef real_point ref_pt_c
    cdef real_point clipping_pt_c
    cdef Polygon *zoi_polygon = NULL

    with nogil:
        # Convert Python to C
        ref_pt_c[XX] = ref_pt[XX]
        ref_pt_c[YY] = ref_pt[YY]

        clipping_pt_c[XX] = clipping_pt[XX]
        clipping_pt_c[YY] = clipping_pt[YY]

        zoi_polygon = polygon_new_from_memview(zoi)

        fast_clip_zoi(zoi_polygon, ref_pt_c, clipping_pt_c, NULL)

    zoi_py = polygon_as_array(zoi_polygon)

    polygon_destroy(zoi_polygon)

    return zoi_py



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function

# Global imports
from unittest import TestCase
import pytest
import numpy
from numpy.testing import assert_almost_equal

# Local imports
from ..core_base import Atom, AtomGroup, PBCBox, Topology


class TestAtom(TestCase):
    def setUp(self):
        self.ref_x = 1.1
        self.ref_y = 2.2
        self.ref_z = 3.3
        self.ref_atomid = 1
        self.ref_name = "H"
        self.atom = Atom(atomid=self.ref_atomid,
                         name=self.ref_name,
                         x=self.ref_x,
                         y=self.ref_y,
                         z=self.ref_z)

    def test_read_atomid(self):
        assert self.atom.atomid == self.ref_atomid

    def test_ro_atomid(self):
        with pytest.raises(AttributeError):
            self.atom.atomid = 0

    def test_read_name(self):
        assert self.atom.name == self.ref_name

    def test_ro_name(self):
        with pytest.raises(AttributeError):
            self.atom.name = "C"

    def test_read_coords(self):
        assert self.atom.x == self.ref_x
        assert self.atom.y == self.ref_y
        assert self.atom.z == self.ref_z

    def test_ro_coords(self):
        with pytest.raises(AttributeError):
            self.atom.x = 0
        with pytest.raises(AttributeError):
            self.atom.y = 0
        with pytest.raises(AttributeError):
            self.atom.z = 0


class TestAtomGroup(TestCase):
    def setUp(self):
        self.atom1 = Atom(x=1)
        self.atom2 = Atom(x=2)
        self.atom3 = Atom(x=3)
        self.atoms = [self.atom1, self.atom2, self.atom3]

    def test_atomgroup_from_atoms(self):
        group = AtomGroup(atoms=self.atoms)
        assert len(group) == len(self.atoms)

        group = AtomGroup(atoms=set(self.atoms))
        assert len(group) == len(self.atoms)

        group = AtomGroup(atoms=tuple(self.atoms))
        assert len(group) == len(self.atoms)

    def test_atomgroup_creation_error(self):
        with pytest.raises(TypeError):
            AtomGroup(atoms="test")

        # iterable but not af atoms, bad!
        with pytest.raises(TypeError):
            AtomGroup(atoms=(self.atom1, None,))

    def test_atomgroup_append(self):
        group = AtomGroup()

        for atom in self.atoms:
            group.append(atom)

        assert len(group) == len(self.atoms)

        with pytest.raises(TypeError):
            group.append(None)


class TestPBCBox(TestCase):
    def setUp(self):
        self.ref_box = numpy.array([[22.25, 0., 0.], [0., 22.25, 0.], [0., 0., 6.36]])
        self.pbcbox = PBCBox(self.ref_box)

        self.coords_raw = numpy.array(
            [[0.47, 12.28, 5.69], [0.11, 12.61, 5.56], [0.2, 12.39, 5.17], [0.33, 12.43, 4.78], [0.47, 12.4, 4.41],
                [0.37, 12.4, 3.99], [0.47, 12.54, 3.6], [-0.02, 12.21, 5.09], [-0.08, 12.21, 4.65], [0.02, 12.19, 4.23],
                [-0.1, 11.95, 3.89], [-0.35, 11.95, 3.48]])

        self.coords_bbox = numpy.array(
            [[0.47, 12.28, 5.69], [0.11, 12.61, 5.56], [0.2, 12.39, 5.17], [0.33, 12.43, 4.78], [0.47, 12.4, 4.41],
                [0.37, 12.4, 3.99], [0.47, 12.54, 3.6], [22.23, 12.21, 5.09], [22.17, 12.21, 4.65], [0.02, 12.19, 4.23],
                [22.15, 11.95, 3.89], [21.90, 11.95, 3.48]])

    def test_pbc_box(self):
        assert_almost_equal(self.pbcbox.box, self.ref_box)

    def test_pbc_asarray(self):
        assert_almost_equal(self.ref_box, self.pbcbox.asarray())

    def test_pbc_put_in_bbox(self):
        coords_bbox = self.pbcbox.put_atoms_in_bbox(self.coords_raw)
        assert_almost_equal(coords_bbox, self.coords_bbox, 2)

    def test_pbc_dx(self):
        ref_dx = self.pbcbox.dx(self.coords_raw[0], self.coords_raw[1])
        dx = self.pbcbox.dx(self.coords_bbox[0], self.coords_bbox[1])

        assert_almost_equal(dx, ref_dx, 2)

    def test_pbc_distance(self):
        ref_dist = self.pbcbox.distance(self.coords_raw[0], self.coords_raw[1])
        dist = self.pbcbox.distance(self.coords_bbox[0], self.coords_bbox[1])

        assert_almost_equal(dist, ref_dist, 2)

    def test_pbc_xcm(self):
        ref_xcm = self.coords_raw.mean(axis=0)
        xcm = self.pbcbox.get_xcm(self.coords_raw)

        assert_almost_equal(xcm, ref_xcm, 2)

    def test_pbc_xcm_consistency(self):
        ref_xcm = self.pbcbox.get_xcm(self.coords_raw)
        xcm = self.pbcbox.get_xcm(self.coords_bbox)

        assert_almost_equal(xcm, ref_xcm, 2)

class TestTopology(TestCase):
    def setUp(self):
        self.topol = Topology()

        self.atoms = [(1, "DPPC", "NC3", 1),
                      (1, "DPPC", "PO4", 2),
                      (1, "DPPC", "GL0", 3),
                      (1, "DPPC", "CA1", 4),
                      (2, "DPPG", "NC3", 5),
                      (2, "DPPG", "PO4", 6),
                      (3, "DPPE ", "NC3", 7)]
        self.ref_atom = self.atoms[1]
        self.residue = ("DPPC1", [1, 2, 3, 4])
        self.resname = "DPPE3"
        self.size = len(self.atoms)

        for atom in self.atoms:
            self.topol.append(*atom)

    def test_size(self):
        assert len(self.topol) == self.size

    def test_atom_access(self):
        assert self.topol.get_atom(2) == self.ref_atom
        assert self.topol.get_atom(2.2) == self.ref_atom

        with pytest.raises(TypeError):
            self.topol.get_atom("1")

        with pytest.raises(KeyError):
            self.topol.get_atom(0)

        with pytest.raises(KeyError):
            self.topol.get_atom(9999)

    def test_residue_access(self):
        residue = self.topol.get_residue_from_atomid(2)
        assert residue["name"] == self.residue[0]
        assert list(residue["atomids"]) == self.residue[1]

        residue = self.topol.get_residue_from_atomid(2.2)
        assert residue["name"] == self.residue[0]
        assert list(residue["atomids"]) == self.residue[1]

        with pytest.raises(TypeError):
            self.topol.get_residue_from_atomid("1")

        with pytest.raises(KeyError):
            self.topol.get_residue_from_atomid(0)

        with pytest.raises(KeyError):
            self.topol.get_residue_from_atomid(9999)

    def test_name_stripping(self):
        assert self.topol.get_residue_from_atomid(7)["name"] == self.resname



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function

# Global imports
import numpy as np
from numpy.testing import assert_almost_equal, assert_allclose

# Local imports
from . import frame_vesicle, frame_model_bilayer_prot, frame_model_bilayer, frame_model_vesicle, frame_bilayer, \
    frame_bilayer_prot, frame_big_prot, frame_bilayer_chol, frame_model_bilayer_vesicle, frame_model_multibilayer, \
    traj_vesicle, frame_bilayer_allatom, frame_multibilayer, frame_bilayer_ganglio


def dprod(v1, v2):
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


def test_openmp():
    from ..core_base import set_num_threads, get_max_threads
    from ..core_analysis import test_parallelism

    nthreads = min(get_max_threads(), 8)

    for i in range(nthreads):
        set_num_threads(i)
        assert test_parallelism()


def test_forcefield_identification_allatom(frame_bilayer_allatom):
    from ..core_base import FORCEFIELDTYPE_ALLATOM
    assert frame_bilayer_allatom.get_forcefield_type() == FORCEFIELDTYPE_ALLATOM


def test_forcefield_identification_unified(frame_bilayer_prot):
    from ..core_base import FORCEFIELDTYPE_UNIFIED
    assert frame_bilayer_prot.get_forcefield_type() == FORCEFIELDTYPE_UNIFIED


def test_forcefield_identification_coarse(frame_bilayer):
    from ..core_base import FORCEFIELDTYPE_COARSE
    assert frame_bilayer.get_forcefield_type() == FORCEFIELDTYPE_COARSE


def test_frame_size(frame_model_bilayer):
    frame = frame_model_bilayer

    assert frame.size == len(frame.bead_coords)


def test_core_base_bbox(frame_vesicle):
    frame = frame_vesicle
    coords = frame.hg_group_coords
    coords_bbox = frame.box.put_atoms_in_bbox(frame.hg_group_coords)

    # All the atoms are already inside the bbox
    assert_almost_equal(coords, coords_bbox)


def test_core_base_interacting_bbox(frame_model_bilayer_prot):
    frame = frame_model_bilayer_prot
    coords = frame.interacting_group_coords_bbox
    coords_bbox = frame.box.put_atoms_in_bbox(frame.interacting_group_coords)

    # All the atoms are already inside the bbox
    assert_almost_equal(coords, coords_bbox)


def test_core_base_directions(frame_model_bilayer):
    frame = frame_model_bilayer

    directions = frame.directions

    ref_direction_first = np.array([0, 0, 1.0])
    ref_direction_last = np.array([0, 0, -1.0])

    assert_almost_equal(directions[0], ref_direction_first, 1)
    assert_almost_equal(directions[-1], ref_direction_last, 1)


def test_core_ns_bilayer(frame_model_bilayer):
    from ..core_ns import neighbor_search

    test_bead = 0
    test_result = sorted([35, 30, 31, 5, 1, 11, 6, 7])

    frame = frame_model_bilayer
    coords = frame.bead_coords

    neighbors = neighbor_search(frame.box, coords, cutoff=1.5)

    assert test_result == sorted(neighbors[test_bead])


def test_core_ns_consistency_bilayer(frame_model_bilayer):
    from ..core_ns import neighbor_search

    frame = frame_model_bilayer
    coords = frame.bead_coords

    neighbors = neighbor_search(frame.box, coords, cutoff=2.0)

    for neighborhood in neighbors:
        assert len(neighborhood) == 20


def test_core_ns_vesicle(frame_vesicle):
    from ..core_ns import neighbor_search

    test_bead = 0
    test_result = sorted([624, 655, 780, 1743, 2104, 1981, 2714, 638, 1352])

    frame = frame_vesicle
    coords = frame.hg_group_coords

    neighbors = neighbor_search(frame.box, coords, cutoff=2)

    assert test_result == sorted(neighbors[test_bead])


def test_core_dx(frame_model_bilayer):
    ref_dx = np.array([0.0, -0.002, 4.336])

    frame = frame_model_bilayer
    coords = frame.bead_coords

    for i in range(36):
        l1_bead = i
        l2_bead = 35 + (i // 6 + 1) * 6 - (i % 6)
        dx = frame.box.dx(coords[l1_bead], coords[l2_bead])

        assert_almost_equal(dx, ref_dx)


def test_directions_bilayer(frame_model_bilayer):
    frame = frame_model_bilayer

    directions = frame.directions

    ref_direction_first = np.array([0, 0, 1.0])
    ref_direction_last = np.array([0, 0, -1.0])

    for i in range(36):
        assert_almost_equal(directions[i], ref_direction_first, 1)
        assert_almost_equal(directions[i + 36], ref_direction_last, 1)


def test_direction_ganglio(frame_bilayer_ganglio):
    frame = frame_bilayer_ganglio

    ref_ganglio = 812
    ref_chol = 27

    assert dprod(frame.directions[ref_chol], [0.0, 0.0, 1.0]) > 0
    assert dprod(frame.directions[ref_ganglio], [0.0, 0.0, 1.0]) > 0
    assert dprod(frame.directions[ref_ganglio], frame.directions[ref_ganglio]) > 0


def test_directions_vesicle(frame_model_vesicle):
    frame = frame_model_vesicle

    directions = frame.directions
    coords = frame.bead_coords
    center = frame.box.center

    for i in range(len(directions)):

        # Get the distance between the lipid and the center
        dx = coords[i] - center
        dx_norm = np.sqrt(np.sum(dx * dx))

        # Change the sign if we are dealing with a lipid from the inner leaflet
        if dx_norm < 7:
            dx *= -1

        dx /= dx_norm

        assert dprod(directions[i], dx) > 0.99


def test_normals_model_bilayer(frame_model_bilayer):
    frame = frame_model_bilayer

    normals = frame.normals

    ref_direction_first = np.array([0, 0, 1.0])
    ref_direction_last = np.array([0, 0, -1.0])

    for i in range(36):
        assert_almost_equal(normals[i], ref_direction_first)
        assert_almost_equal(normals[i + 36], ref_direction_last)


def test_normals_bilayer(frame_bilayer):
    frame = frame_bilayer

    normals = frame.normals

    coords = frame.bead_coords
    center_z = 3.5

    ref_direction_up = np.array([0, 0, 1.0])
    ref_direction_down = np.array([0, 0, -1.0])

    for i, normal in enumerate(normals):
        if coords[i][2] > center_z:
            assert dprod(normal, ref_direction_up) > 0.97
        else:
            assert dprod(normal, ref_direction_down) > 0.97


def test_normals_bilayer_prot(frame_bilayer_prot):
    frame = frame_bilayer_prot

    normals = frame.normals
    coords = frame.bead_coords
    center_z = 3.5

    ref_direction_up = np.array([0, 0, 1.0])
    ref_direction_down = np.array([0, 0, -1.0])

    for i, normal in enumerate(normals):
        if coords[i][2] > center_z:
            assert dprod(normal, ref_direction_up) > 0.95
        else:
            assert dprod(normal, ref_direction_down) > 0.95


def test_normals_vesicle_model(frame_model_vesicle):
    frame = frame_model_vesicle

    normals = frame.normals
    coords = frame.bead_coords
    center = frame.box.center

    for i in range(len(normals)):

        # Get the distance between the lipid and the center
        dx = coords[i] - center
        dx_norm = np.sqrt(np.sum(dx * dx))

        # Change the sign if we are dealing with a lipid from the inner leaflet
        if dx_norm < 7:
            dx *= -1

        dx /= dx_norm

        assert dprod(normals[i], dx) > 0.99


def test_normals_big_prot(frame_big_prot):
    test_bead = 6876
    test_neighbors = [21465, 6975, 8575, 8994, 12676, 8731, 8803, 10220, 10573, 12151, 12363, 12366,
                      22388, 6996, 8673, 8786, 8842, 8990, 12283, 12403, 22284, 10471, 21441]
    test_bad = [21441, 10471, 8731]
    frame = frame_big_prot

    normals = frame.normals

    for nid in test_neighbors:
        dp_val = dprod(normals[test_bead], normals[nid])

        if nid in test_bad:
            assert dp_val < 0
        else:
            assert dp_val > 0.90


def test_core_dx_leaflet(frame_model_bilayer):
    ref_dx = np.array([0.000, -0.002, -5.664])

    frame = frame_model_bilayer
    coords = frame.bead_coords
    normals = frame.normals

    for i in range(36):
        l1_bead = i
        l2_bead = 35 + (i // 6 + 1) * 6 - (i % 6)

        dx = frame.box.dx_leaflet(coords[l1_bead], coords[l2_bead], normals[l1_bead])
        assert_almost_equal(dx, ref_dx)

        dx = frame.box.dx_leaflet(coords[l2_bead], coords[l1_bead], normals[l2_bead])
        assert_almost_equal(dx, -ref_dx)


def test_aggregate_model_bilayer(frame_model_bilayer):
    ref_aggregates = [np.arange(36), np.arange(36, 72)]
    frame = frame_model_bilayer
    aggregates = frame.get_aggregates()

    assert len(aggregates) == len(ref_aggregates)

    for i, aggregate in enumerate(aggregates):
        assert_almost_equal(aggregate.beadids, ref_aggregates[i])
        assert_allclose(aggregate.xcm, aggregate.coords.mean(axis=0), 1e-2)

    assert len(frame.get_aggregates(0.5)) == 72


def test_aggregate_bilayer(frame_bilayer):
    ref_aggregates = [169, 169]

    frame = frame_bilayer
    aggregates = frame.get_aggregates()

    assert len(aggregates) == len(ref_aggregates)

    for i, aggregate in enumerate(aggregates):
        assert len(aggregate) == ref_aggregates[i]


def test_aggregate_big_prot(frame_big_prot):
    ref_aggregates = [12152, 11861, 42, 1]
    ref_xcm = [
        [43.19, 43.06, 9.95],
        [43.24, 43.20, 14.27]
    ]
    expected_normal = [
        [0, 0, -1],
        [0, 0, 1]
    ]

    frame = frame_big_prot
    aggregates = frame.get_aggregates()

    assert len(aggregates) == len(ref_aggregates)

    for i, aggregate in enumerate(aggregates):
        assert len(aggregate) == ref_aggregates[i]

        if i < len(ref_xcm):
            assert_allclose(aggregate.xcm, ref_xcm[i], rtol=0.001)

            dp_val = dprod(aggregate.avg_normal, expected_normal[i])

            assert dp_val > 0.90


def test_aggregate_chol(frame_bilayer_chol):
    ref_aggregates = [977, 967]

    frame = frame_bilayer_chol
    aggregates = frame.get_aggregates()

    assert len(aggregates) == len(ref_aggregates)

    for i, aggregate in enumerate(aggregates):
        assert len(aggregate) == ref_aggregates[i]


def test_aggregate_vesicle_model(frame_model_vesicle):
    frame = frame_model_vesicle
    aggregates = frame.get_aggregates()

    ref_aggregates = [1963, 785]

    assert len(aggregates) == len(ref_aggregates)

    for i, aggregate in enumerate(aggregates):
        assert len(aggregate) == ref_aggregates[i]


def test_aggregate_vesicle(frame_vesicle):
    frame = frame_vesicle
    aggregates = frame.get_aggregates()

    ref_aggregates = [1851, 1179, 26, 3, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1]

    assert len(aggregates) == len(ref_aggregates)

    for i, aggregate in enumerate(aggregates):
        assert len(aggregate) == ref_aggregates[i]
        if len(aggregates) > 100:
            assert_allclose(aggregate.xcm, (26.9,  36.7,  15.4), 1e-2)


def test_aggregate_vesicle_traj(traj_vesicle):
    nlipids = len(traj_vesicle[0].bead_coords)

    for frame in traj_vesicle:
        aggregates = frame.get_aggregates()
        assert sum([len(val) for val in aggregates]) == nlipids


def test_aggregate_atomid(frame_model_bilayer):
    ref_atomids = (4, 1804)

    frame = frame_model_bilayer
    aggregates = frame.get_aggregates()

    assert aggregates[0].get_atomid(0) == ref_atomids[0]
    assert aggregates[1].get_atomid(0) == ref_atomids[1]


def test_aggregate_residue(frame_model_bilayer):
    ref_atomid = 1804

    frame = frame_model_bilayer
    aggregates = frame.get_aggregates()
    ref_residue = frame.trajectory.get_topology().get_residue_from_atomid(ref_atomid)

    for key, val in aggregates[1].get_residue(0).items():
        try:
            assert ref_residue[key] == val
        except ValueError:
            assert_almost_equal(ref_residue[key], val)


def test_aggregate_resname(frame_model_bilayer):
    ref_atomid = 1804
    ref_resname = "DPPC"

    frame = frame_model_bilayer
    aggregates = frame.get_aggregates()

    assert aggregates[1].get_resname(0) == ref_resname
    assert aggregates[1].get_resname(0) == frame.trajectory.\
        get_topology().get_residue_from_atomid(ref_atomid)["resname"]


def test_aggregate_same_restype(frame_bilayer_chol):
    frame = frame_bilayer_chol
    aggregates = frame.get_aggregates()

    assert aggregates[0].same_restype(0, 1)
    assert not aggregates[0].same_restype(0, len(aggregates[0])-1)


def test_membranes_model_bilayer(frame_model_bilayer):
    frame = frame_model_bilayer
    nlipids = frame.size

    ref_membrane_beadids = [np.arange(36, 72),
                            np.arange(36)]
    membranes = frame.get_membranes()

    assert len(membranes) == 1
    membrane = membranes[0]

    assert membrane.is_planar

    assert sum([len(val) for val in membrane]) == nlipids

    for i, beadids in enumerate(membrane.beadids):
        assert_almost_equal(beadids, ref_membrane_beadids[i])


def test_membranes_bilayer(frame_bilayer):
    ref_leaflets = [169, 169]

    frame = frame_bilayer
    membranes = frame.get_membranes()

    assert len(membranes) == 1

    membrane = membranes[0]

    assert membrane.is_planar

    for i, leaflet in enumerate(membrane):
        assert len(leaflet) == ref_leaflets[i]


def test_membranes_bilayer_normals(frame_bilayer):
    frame = frame_bilayer

    membrane = frame.get_membranes()[0]

    center_z = 3.5

    ref_direction_up = np.array([0, 0, 1.0])
    ref_direction_down = np.array([0, 0, -1.0])

    for leaflet in membrane:
        coords = leaflet.coords
        normals = leaflet.normals

        for i, normal in enumerate(normals):
            if coords[i][2] > center_z:
                assert dprod(normal, ref_direction_up) > 0.985  # less than 10° off
            else:
                assert dprod(normal, ref_direction_down) > 0.985  # less than 10° off


def test_membranes_big_prot(frame_big_prot):
    ref_leaflets = [12152, 11861]

    frame = frame_big_prot
    membranes = frame.get_membranes()

    assert len(membranes) == 1

    membrane = membranes[0]

    assert membrane.is_planar

    for i, leaflet in enumerate(membrane):
        assert len(leaflet) == ref_leaflets[i]


def test_membranes_chol(frame_bilayer_chol):

    ref_leaflets = [977, 967]

    frame = frame_bilayer_chol
    membranes = frame.get_membranes()

    assert len(membranes) == 1

    membrane = membranes[0]

    assert membrane.is_planar

    for i, leaflet in enumerate(membrane):
        assert len(leaflet) == ref_leaflets[i]


def test_membranes_vesicle_model(frame_model_vesicle):
    ref_leaflets = [1963, 785]

    frame = frame_model_vesicle
    membranes = frame.get_membranes()

    assert len(membranes) == 1

    membrane = membranes[0]

    assert membrane.is_vesicle

    for i, leaflet in enumerate(membrane):
        assert len(leaflet) == ref_leaflets[i]


def test_membranes_vesicle_model_normals(frame_model_vesicle):
    frame = frame_model_vesicle
    membrane = frame.get_membranes()[0]

    for lid, leaflet in enumerate(membrane):
        coords = leaflet.coords
        normals = leaflet.normals
        xcm = leaflet.xcm

        for i, normal in enumerate(normals):
            ref_direction = coords[i] - xcm
            ref_direction /= np.linalg.norm(ref_direction)
            if lid == 1:
                ref_direction *= -1.0
            assert dprod(normal, ref_direction) > 0.998  # less than 4° off


def test_membranes_vesicle(frame_vesicle):
    ref_leaflets = [1851, 1179]

    frame = frame_vesicle
    membranes = frame.get_membranes()

    assert len(membranes) == 1

    membrane = membranes[0]

    assert membrane.is_vesicle

    for i, leaflet in enumerate(membrane):
        assert len(leaflet) == ref_leaflets[i]


def test_membranes_vesicle_traj(traj_vesicle):
    for frame in traj_vesicle:
        membranes = frame.get_membranes()
        assert len(membranes) == 1
        assert membranes[0].is_vesicle


def test_membranes_multibilayer_model(frame_model_multibilayer):

    frame = frame_model_multibilayer
    nlipids = 72

    ref_membrane_beadids = [[np.arange(36, 72),
                            np.arange(36)],
                            [np.arange(72, 108),
                                np.arange(108, 144)]]
    membranes = frame.get_membranes()

    assert len(membranes) == 2

    for mid, membrane in enumerate(membranes):
        assert membrane.is_planar

        assert sum([len(val) for val in membrane]) == nlipids

        for i, beadids in enumerate(membrane.beadids):
            assert_almost_equal(beadids, ref_membrane_beadids[mid][i])


def test_membranes_multibilayer(frame_multibilayer):

    frame = frame_multibilayer
    nlipids = 256

    ref_membrane_beadids_firsts = [
        [
            [198, 240, 254, 366, 506],
            [30, 44, 58, 86, 142]
        ],
        [
            [2, 16, 72, 100, 114],
            [688, 1682, 2022, 2036, 2050]
        ]
    ]

    membranes = frame.get_membranes()

    assert len(membranes) == 2

    for mid, membrane in enumerate(membranes):
        assert membrane.is_planar

        assert sum([len(val) for val in membrane]) == nlipids

        for i, leaflet in enumerate(membrane):
            assert_almost_equal(leaflet.hg_atomids[:5], ref_membrane_beadids_firsts[mid][i])


def test_membranes_bilayer_vesicle(frame_model_bilayer_vesicle):
    frame = frame_model_bilayer_vesicle

    ref_leaflets = [[1963, 785], [1296, 1296]]
    membrane_planar = [False, True]

    membranes = frame.get_membranes()

    assert len(membranes) == 2

    for mid, membrane in enumerate(membranes):

        assert membrane.is_planar == membrane_planar[mid]

        assert membrane.is_vesicle != membrane_planar[mid]

        for i, leaflet in enumerate(membrane):
            assert len(leaflet) == ref_leaflets[mid][i]


def test_membranes_bilayer_ganglio(frame_bilayer_ganglio):
    frame = frame_bilayer_ganglio

    membrane = frame.get_membranes()[0]

    assert len(membrane) == len(frame.bead_coords)



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function

# Global imports
from unittest import TestCase

import pytest
import numpy
from numpy.testing import assert_almost_equal


# Local imports
from .data import MODEL_BILAYER_GRO, MODEL_BILAYER_NDX, VESICLE_GRO, VESICLE_NDX, VESICLE_XTC, \
    VESICLE_TRR, \
    MODEL_BIG_GRO, MODEL_BIG_NDX, MODEL_BILAYER_PROT_GRO, MODEL_BILAYER_PROT_NDX, \
    BILAYER_ALLATOM_GRO, BILAYER_ALLATOM_NDX, BILAYER_PEPTIDE_GRO, BILAYER_PEPTIDE_H_NDX, \
    BILAYER_CHOL_GRO, BILAYER_CHOL_NDX
from ..datareading import UnknownFileType, get_readers, load_trajectory
from ..core_datareading import NdxReader


def test_index_filetype():
    with pytest.raises(ValueError):
        NdxReader("fake.index")


class TestIndexReader(TestCase):
    def setUp(self):
        self.index = NdxReader(MODEL_BILAYER_NDX)
        self.ref_size = 4
        self.ref_group = "headgroups"
        self.ref_group_size = 144
        self.ref_group_firsts = numpy.array([4, 8, 54, 58, 104], dtype=int)
        self.ref_group_lasts = numpy.array([3458, 3504, 3508, 3554, 3558], dtype=int)

    def test_size(self):
        assert len(self.index) == self.ref_size

    def test_group_size(self):
        assert len(self.index[self.ref_group]) == self.ref_group_size

    def test_group_firsts(self):
        assert_almost_equal(self.index[self.ref_group][:5], self.ref_group_firsts)

    def test_group_lasts(self):
        assert_almost_equal(self.index[self.ref_group][-5:], self.ref_group_lasts)

    def test_fake_group(self):
        with pytest.raises(KeyError):
            assert self.index["fake_group"]


def test_unk_reader():
    with pytest.raises(UnknownFileType):
        get_readers("test.unk", MODEL_BILAYER_NDX)

    with pytest.raises(UnknownFileType):
        get_readers(MODEL_BILAYER_GRO, "test.unk")

    with pytest.raises(UnknownFileType):
        get_readers(MODEL_BILAYER_GRO, MODEL_BILAYER_NDX, MODEL_BILAYER_NDX)

    # This should not raise anything
    get_readers(MODEL_BILAYER_GRO, MODEL_BILAYER_NDX, "test.unk")


def test_gro_reader():
    assert get_readers(MODEL_BILAYER_GRO, MODEL_BILAYER_NDX)


def test_trajectory_initialization():
    traj = load_trajectory(MODEL_BILAYER_GRO, MODEL_BILAYER_NDX)

    with pytest.raises(ValueError):
        assert traj[0]


def test_trajectory_initialization_hydrogen():
    traj = load_trajectory(BILAYER_PEPTIDE_GRO, BILAYER_PEPTIDE_H_NDX)
    with pytest.raises(KeyError):
        traj.initialize()


class TestTopolReading(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.traj = load_trajectory(BILAYER_ALLATOM_GRO, BILAYER_ALLATOM_NDX)
        cls.traj.initialize()
        cls.topology = cls.traj[0].topology

    def test_hydrogens_skipped(self):
        assert self.topology.get_residue(1)["size"] == 46

    def test_solvent_skipped(self):
        assert self.topology.natoms == 128 * 46
        assert self.topology.nresidues == 128


class TestGroReading(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.traj = load_trajectory(MODEL_BILAYER_PROT_GRO, MODEL_BILAYER_PROT_NDX)
        cls.traj.initialize()

    def test_hg_atomids(self):
        frame = self.traj[0]
        atomids = frame.hg_atomids

        ref_group_firsts = numpy.array([2227, 2231, 2277, 2281, 2327], dtype=int)
        ref_group_lasts = numpy.array([26081, 26127, 26131, 26177, 26181], dtype=int)

        assert_almost_equal(atomids[:5], ref_group_firsts)
        assert_almost_equal(atomids[-5:], ref_group_lasts)

    def test_lipid_atomids(self):
        atomids = self.traj[0].lipid_atomids
        ref_group_firsts = numpy.array([2224, 2225, 2226, 2227, 2228], dtype=int)
        ref_group_lasts = numpy.array([26219, 26220, 26221, 26222, 26223], dtype=int)

        assert_almost_equal(atomids[0][:5], ref_group_firsts)
        assert_almost_equal(atomids[-1][-5:], ref_group_lasts)

    def test_timesteps(self):
        assert self.traj[0].timestep == 0.0

    def test_traj_size(self):
        assert len(self.traj) == 1

    def test_pbcbox(self):
        assert_almost_equal(self.traj[0].box.asarray(), numpy.array([[12.8, 0, 0], [0, 12.8, 0], [0, 0, 10.0]]))

    def test_hg_coords(self):
        coords = self.traj[0].hg_group_coords

        ref_coords_first = numpy.array((0.362, 0.356, 8.039))
        ref_coords_last = numpy.array((12.385, 11.554, 2.375))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    def test_lipid_coords(self):
        coords = self.traj[0].lipid_coords

        ref_coords_first = numpy.array((0.459, 0.281, 8.119))
        ref_coords_last = numpy.array((12.372, 11.480, 4.924))

        assert_almost_equal(coords[0][0], ref_coords_first)
        assert_almost_equal(coords[-1][-1], ref_coords_last)

    def test_bead_coords(self):
        coords = self.traj[0].bead_coords

        ref_coords_first = 0.5 * (numpy.array((0.362, 0.356, 8.039)) + numpy.array((0.385, 0.446, 7.625)))
        ref_coords_last = 0.5 * (numpy.array((12.362, 11.644, 1.961)) + numpy.array((12.385, 11.554, 2.375)))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    def test_interacting_coords(self):
        coords = self.traj[0].interacting_group_coords

        assert len(coords) > 0

        ref_coords_first = numpy.array((7.258, 6.204, 7.468))
        ref_coords_last = numpy.array((7.312, 5.275, 6.634))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    @classmethod
    def tearDownClass(cls):
        del cls.traj


class TestBigGroReading(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.traj = load_trajectory(MODEL_BIG_GRO, MODEL_BIG_NDX)
        cls.traj.initialize()

    def test_hg_atomids(self):
        atomids = self.traj[0].hg_atomids

        ref_group_firsts = numpy.array([8, 58, 108, 158, 208], dtype=int)
        ref_group_lasts = numpy.array([129358, 129408, 129458, 129508, 129558], dtype=int)

        assert_almost_equal(atomids[:5], ref_group_firsts)
        assert_almost_equal(atomids[-5:], ref_group_lasts)

    def test_lipid_atomids(self):
        atomids = self.traj[0].lipid_atomids
        ref_group_firsts = numpy.array([1, 2, 3, 4, 5], dtype=int)
        ref_group_lasts = numpy.array([129596, 129597, 129598, 129599, 129600], dtype=int)

        assert_almost_equal(atomids[0][:5], ref_group_firsts)
        assert_almost_equal(atomids[-1][-5:], ref_group_lasts)

    def test_timesteps(self):
        assert self.traj[0].timestep == 0.0

    def test_traj_size(self):
        assert len(self.traj) == 1

    def test_pbcbox(self):
        assert_almost_equal(self.traj[0].box.asarray(), numpy.array([[28.8, 0, 0], [0, 28.8, 0], [0, 0, 10.0]]))

    def test_hg_coords(self):
        coords = self.traj[0].hg_group_coords

        ref_coords_first = numpy.array((0.385, 0.446, 7.625))
        ref_coords_last = numpy.array((28.385, 24.354, 2.375))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    def test_lipid_coords(self):
        coords = self.traj[0].lipid_coords

        ref_coords_first = numpy.array((0.459, 0.281, 8.119))
        ref_coords_last = numpy.array((28.372, 24.280, 4.924))

        assert_almost_equal(coords[0][0], ref_coords_first)
        assert_almost_equal(coords[-1][-1], ref_coords_last)

    def test_bead_coords(self):
        coords = self.traj[0].bead_coords

        ref_coords_first = numpy.array((0.385, 0.446, 7.625))
        ref_coords_last = numpy.array((28.385, 24.354, 2.375))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    @classmethod
    def tearDownClass(cls):
        del cls.traj


class TestCholReading(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.traj = load_trajectory(BILAYER_CHOL_GRO, BILAYER_CHOL_NDX)
        cls.traj.initialize()

    def test_hg_atomids(self):
        atomids = self.traj[0].hg_atomids

        ref_group_firsts = numpy.array([2, 14, 26, 38, 50], dtype=int)
        ref_group_lasts = numpy.array([33585, 33593, 33601, 33609, 33617], dtype=int)

        assert_almost_equal(atomids[:5], ref_group_firsts)
        assert_almost_equal(atomids[-5:], ref_group_lasts)

    def test_lipid_atomids(self):
        atomids = self.traj[0].lipid_atomids
        ref_group_firsts = numpy.array([1, 2, 3, 4, 5], dtype=int)
        ref_group_lasts = numpy.array([33620, 33621, 33622, 33623, 33624], dtype=int)

        assert_almost_equal(atomids[0][:5], ref_group_firsts)
        assert_almost_equal(atomids[-1][-5:], ref_group_lasts)

    def test_timesteps(self):
        assert self.traj[0].timestep == 0.0

    def test_traj_size(self):
        assert len(self.traj) == 1

    def test_pbcbox(self):
        assert_almost_equal(self.traj[0].box.asarray(), numpy.array([[21.6588, 0, 0],
                                                                     [0, 21.82714, 0],
                                                                     [0, 0, 7.48249]]))

    def test_hg_coords(self):
        coords = self.traj[0].hg_group_coords

        ref_coords_first = numpy.array((11.950, 9.160, 1.680))
        ref_coords_last = numpy.array((16.720, 0.350, 4.450))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    def test_lipid_coords(self):
        coords = self.traj[0].lipid_coords

        ref_coords_first = numpy.array((12.300, 9.160, 1.530))
        ref_coords_last = numpy.array((16.590, 0.120, 3.030))

        assert_almost_equal(coords[0][0], ref_coords_first)
        assert_almost_equal(coords[-1][-1], ref_coords_last)

    def test_bead_coords(self):
        coords = self.traj[0].bead_coords

        ref_coords_first = numpy.array((11.950, 9.160, 1.680))
        ref_coords_last = numpy.array((16.720, 0.350, 4.450))

        assert_almost_equal(coords[0], ref_coords_first)
        assert_almost_equal(coords[-1], ref_coords_last)

    @classmethod
    def tearDownClass(cls):
        del cls.traj


class TestXtcReading(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.traj = load_trajectory(VESICLE_GRO, VESICLE_NDX, VESICLE_XTC)
        cls.traj.initialize()

    @classmethod
    def tearDownClass(cls):
        del cls.traj

    def test_hg_atomids(self):
        atomids = self.traj[0].hg_atomids

        ref_group_firsts = numpy.array([2, 14, 26, 38, 50], dtype=int)
        ref_group_lasts = numpy.array([36806, 36818, 36830, 36842, 36854], dtype=int)

        assert_almost_equal(atomids[:5], ref_group_firsts)
        assert_almost_equal(atomids[-5:], ref_group_lasts)

    def test_lipid_atomids(self):
        atomids = self.traj[0].lipid_atomids

        ref_group_firsts = numpy.array([1, 2, 3, 4, 5], dtype=int)
        ref_group_lasts = numpy.array([36860, 36861, 36862, 36863, 36864], dtype=int)

        assert_almost_equal(atomids[0][:5], ref_group_firsts)
        assert_almost_equal(atomids[-1][-5:], ref_group_lasts)

    def test_traj_size(self):
        assert len(self.traj) == 11

    def test_timesteps(self):
        ref_timesteps = [500.00 * i for i in range(len(self.traj))]

        for i, frame in enumerate(self.traj):
            assert frame.timestep == ref_timesteps[i]

    def test_pbcbox(self):
        ref_boxes = [numpy.array([[47.100067138671875, 0.0, 0.0], [0.0, 47.29520797729492, 0.0],
            [23.550033569335938, 23.64760398864746, 31.985841751098633]]),
            numpy.array([[47.08340835571289, 0.0, 0.0], [0.0, 47.30361557006836, 0.0],
                [23.541704177856445, 23.65180778503418, 31.991649627685547]]),
            numpy.array([[47.08655548095703, 0.0, 0.0], [0.0, 47.298133850097656, 0.0],
                [23.543277740478516, 23.649066925048828, 31.99555778503418]]),
            numpy.array([[47.07117462158203, 0.0, 0.0], [0.0, 47.30364227294922, 0.0],
                [23.535587310791016, 23.65182113647461, 31.99390983581543]]),
            numpy.array([[47.079769134521484, 0.0, 0.0], [0.0, 47.30522918701172, 0.0],
                [23.539884567260742, 23.65261459350586, 31.995153427124023]]),
            numpy.array([[47.07292175292969, 0.0, 0.0], [0.0, 47.30583953857422, 0.0],
                [23.536460876464844, 23.65291976928711, 32.00019073486328]]),
            numpy.array([[47.066261291503906, 0.0, 0.0], [0.0, 47.3122444152832, 0.0],
                [23.533130645751953, 23.6561222076416, 31.99447250366211]]),
            numpy.array([[47.06310272216797, 0.0, 0.0], [0.0, 47.31338119506836, 0.0],
                [23.531551361083984, 23.65669059753418, 31.999902725219727]]),
            numpy.array([[47.075565338134766, 0.0, 0.0], [0.0, 47.29836654663086, 0.0],
                [23.537782669067383, 23.64918327331543, 32.0003662109375]]),
            numpy.array([[47.0737419128418, 0.0, 0.0], [0.0, 47.295833587646484, 0.0],
                [23.5368709564209, 23.647916793823242, 32.00419235229492]]),
            numpy.array([[47.06859588623047, 0.0, 0.0], [0.0, 47.29644775390625, 0.0],
                [23.534297943115234, 23.648223876953125, 32.00706100463867]]),
        ]

        for i, frame in enumerate(self.traj):
            assert_almost_equal(frame.box.asarray(), ref_boxes[i], 5)

    def test_hg_coords(self):
        ref_coords = [(numpy.array([46.704, 26.634, 25.838]), numpy.array([0.828, 25.352, 23.336])),
            (numpy.array([0.245, 25.697, 24.964]), numpy.array([1.076, 25.148, 23.05])),
            (numpy.array([46.654, 25.595, 24.466]), numpy.array([0.925, 24.454, 21.854])),
            (numpy.array([0.452, 26.92, 24.795]), numpy.array([1.798, 25.228, 21.268])),
            (numpy.array([0.153, 25.762, 24.212]), numpy.array([1.903, 24.785, 21.654])),
            (numpy.array([0.705, 26.689, 24.573]), numpy.array([1.880, 24.884, 21.318])),
            (numpy.array([0.278, 27.591, 25.560]), numpy.array([1.44, 25.025, 20.538])),
            (numpy.array([0.244, 27.5, 24.755]), numpy.array([1.751, 25.319, 20.476])),
            (numpy.array([0.604, 27.412, 24.827]), numpy.array([1.290, 24.705, 19.278])),
            (numpy.array([46.885, 27.279, 24.879]), numpy.array([1.415, 24.76, 19.318])),
            (numpy.array([47.049, 27.8, 25.764]), numpy.array([1.561, 25.331, 19.281])), ]

        for i, frame in enumerate(self.traj):
            ref_coords_first, ref_coords_last = ref_coords[i]
            coords = frame.hg_group_coords

            assert_almost_equal(coords[0], ref_coords_first, 3)
            assert_almost_equal(coords[-1], ref_coords_last, 3)

    def test_lipid_coords(self):
        ref_coords = [(numpy.array([47.089, 26.709, 25.909]), numpy.array([46.848, 26.583, 21.697])),
            (numpy.array([46.983, 25.571, 24.778]), numpy.array([46.677, 26.159, 21.724])),
            (numpy.array([46.917, 25.729, 24.715]), numpy.array([46.616, 26.104, 21.236])),
            (numpy.array([0.911, 26.731, 24.829]), numpy.array([0.292, 26.295, 21.759])),
            (numpy.array([46.834, 25.704, 24.316]), numpy.array([1.705, 26.92, 20.982])),
            (numpy.array([0.976, 26.966, 24.42]), numpy.array([0.986, 26.455, 20.507])),
            (numpy.array([47.02, 27.213, 25.534]), numpy.array([0.909, 26.701, 19.69])),
            (numpy.array([0.622, 27.231, 24.827]), numpy.array([1.296, 27.084, 19.463])),
            (numpy.array([0.792, 27.017, 24.68]), numpy.array([0.342, 26.490, 19.179])),
            (numpy.array([47.022, 27.047, 25.295]), numpy.array([2.073, 26.168, 19.01])),
            (numpy.array([0.3, 27.619, 25.796]), numpy.array([1.17, 27.371, 19.19])), ]

        for i, frame in enumerate(self.traj):
            ref_coords_first, ref_coords_last = ref_coords[i]
            coords = frame.lipid_coords

            assert_almost_equal(coords[0][0], ref_coords_first, 3)
            assert_almost_equal(coords[-1][-1], ref_coords_last, 3)

    def test_bead_coords(self):
        ref_coords = [(numpy.array([46.704, 26.634, 25.838]), numpy.array([0.828, 25.352, 23.336])),
            (numpy.array([0.245, 25.697, 24.964]), numpy.array([1.076, 25.148, 23.05])),
            (numpy.array([46.654, 25.595, 24.466]), numpy.array([0.925, 24.454, 21.854])),
            (numpy.array([0.452, 26.92, 24.795]), numpy.array([1.798, 25.228, 21.268])),
            (numpy.array([0.153, 25.762, 24.212]), numpy.array([1.903, 24.785, 21.654])),
            (numpy.array([0.705, 26.689, 24.573]), numpy.array([1.880, 24.884, 21.318])),
            (numpy.array([0.278, 27.591, 25.560]), numpy.array([1.44, 25.025, 20.538])),
            (numpy.array([0.244, 27.5, 24.755]), numpy.array([1.751, 25.319, 20.476])),
            (numpy.array([0.604, 27.412, 24.827]), numpy.array([1.290, 24.705, 19.278])),
            (numpy.array([46.885, 27.279, 24.879]), numpy.array([1.415, 24.76, 19.318])),
            (numpy.array([47.049, 27.8, 25.764]), numpy.array([1.561, 25.331, 19.281])), ]

        for i, frame in enumerate(self.traj):
            ref_coords_first, ref_coords_last = ref_coords[i]
            coords = frame.bead_coords

            assert_almost_equal(coords[0], ref_coords_first, 3)
            assert_almost_equal(coords[-1], ref_coords_last, 3)


class TestTrrReading(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.traj = load_trajectory(VESICLE_GRO, VESICLE_NDX, VESICLE_TRR)
        cls.traj.initialize()

    @classmethod
    def tearDownClass(cls):
        del cls.traj

    def test_hg_atomids(self):
        atomids = self.traj[0].hg_atomids

        ref_group_firsts = numpy.array([2, 14, 26, 38, 50], dtype=int)
        ref_group_lasts = numpy.array([36806, 36818, 36830, 36842, 36854], dtype=int)

        assert_almost_equal(atomids[:5], ref_group_firsts)
        assert_almost_equal(atomids[-5:], ref_group_lasts)

    def test_lipid_atomids(self):
        atomids = self.traj[0].lipid_atomids

        ref_group_firsts = numpy.array([1, 2, 3, 4, 5], dtype=int)
        ref_group_lasts = numpy.array([36860, 36861, 36862, 36863, 36864], dtype=int)

        assert_almost_equal(atomids[0][:5], ref_group_firsts)
        assert_almost_equal(atomids[-1][-5:], ref_group_lasts)

    def test_traj_size(self):
        assert len(self.traj) == 11

    def test_timesteps(self):
        ref_timesteps = [500.00 * i for i in range(len(self.traj))]

        for i, frame in enumerate(self.traj):
            assert frame.timestep == ref_timesteps[i]

    def test_pbcbox(self):
        ref_boxes = [numpy.array([[47.100067138671875, 0.0, 0.0], [0.0, 47.29520797729492, 0.0],
            [23.550033569335938, 23.64760398864746, 31.985841751098633]]),
            numpy.array([[47.08340835571289, 0.0, 0.0], [0.0, 47.30361557006836, 0.0],
                [23.541704177856445, 23.65180778503418, 31.991649627685547]]),
            numpy.array([[47.08655548095703, 0.0, 0.0], [0.0, 47.298133850097656, 0.0],
                [23.543277740478516, 23.649066925048828, 31.99555778503418]]),
            numpy.array([[47.07117462158203, 0.0, 0.0], [0.0, 47.30364227294922, 0.0],
                [23.535587310791016, 23.65182113647461, 31.99390983581543]]),
            numpy.array([[47.079769134521484, 0.0, 0.0], [0.0, 47.30522918701172, 0.0],
                [23.539884567260742, 23.65261459350586, 31.995153427124023]]),
            numpy.array([[47.07292175292969, 0.0, 0.0], [0.0, 47.30583953857422, 0.0],
                [23.536460876464844, 23.65291976928711, 32.00019073486328]]),
            numpy.array([[47.066261291503906, 0.0, 0.0], [0.0, 47.3122444152832, 0.0],
                [23.533130645751953, 23.6561222076416, 31.99447250366211]]),
            numpy.array([[47.06310272216797, 0.0, 0.0], [0.0, 47.31338119506836, 0.0],
                [23.531551361083984, 23.65669059753418, 31.999902725219727]]),
            numpy.array([[47.075565338134766, 0.0, 0.0], [0.0, 47.29836654663086, 0.0],
                [23.537782669067383, 23.64918327331543, 32.0003662109375]]),
            numpy.array([[47.0737419128418, 0.0, 0.0], [0.0, 47.295833587646484, 0.0],
                [23.5368709564209, 23.647916793823242, 32.00419235229492]]),
            numpy.array([[47.06859588623047, 0.0, 0.0], [0.0, 47.29644775390625, 0.0],
                [23.534297943115234, 23.648223876953125, 32.00706100463867]]),
        ]

        for i, frame in enumerate(self.traj):
            assert_almost_equal(frame.box.asarray(), ref_boxes[i], 5)

    def test_hg_coords(self):
        ref_coords = [(numpy.array([36.704, 26.634, 25.838]), numpy.array([37.928, 25.352, 23.336])),
            (numpy.array([37.328, 25.697, 24.964]), numpy.array([38.159, 25.148, 23.05])),
            (numpy.array([36.654, 25.595, 24.466]), numpy.array([38.012, 24.454, 21.854])),
            (numpy.array([37.523, 26.92, 24.795]), numpy.array([38.869, 25.228, 21.268])),
            (numpy.array([37.233, 25.762, 24.212]), numpy.array([38.983, 24.785, 21.654])),
            (numpy.array([37.778, 26.689, 24.573]), numpy.array([38.953, 24.884, 21.318])),
            (numpy.array([37.344, 27.591, 25.56]), numpy.array([38.506, 25.025, 20.538])),
            (numpy.array([37.307, 27.5, 24.755]), numpy.array([38.814, 25.319, 20.476])),
            (numpy.array([37.68, 27.412, 24.827]), numpy.array([38.366, 24.705, 19.278])),
            (numpy.array([36.885, 27.279, 24.879]), numpy.array([38.489, 24.76, 19.318])),
            (numpy.array([37.049, 27.8, 25.764]), numpy.array([38.63, 25.331, 19.281])), ]

        for i, frame in enumerate(self.traj):
            ref_coords_first, ref_coords_last = ref_coords[i]
            coords = frame.hg_group_coords

            assert_almost_equal(coords[0], ref_coords_first, 3)
            assert_almost_equal(coords[-1], ref_coords_last, 3)

    def test_lipid_coords(self):
        ref_coords = [(numpy.array([37.089, 26.709, 25.909]), numpy.array([36.848, 26.583, 21.697])),
            (numpy.array([36.983, 25.571, 24.778]), numpy.array([36.677, 26.159, 21.724])),
            (numpy.array([36.917, 25.729, 24.715]), numpy.array([36.616, 26.104, 21.236])),
            (numpy.array([37.982, 26.731, 24.829]), numpy.array([37.363, 26.295, 21.759])),
            (numpy.array([36.834, 25.704, 24.316]), numpy.array([38.785, 26.92, 20.982])),
            (numpy.array([38.049, 26.966, 24.42]), numpy.array([38.059, 26.455, 20.507])),
            (numpy.array([37.02, 27.213, 25.534]), numpy.array([37.975, 26.701, 19.69])),
            (numpy.array([37.685, 27.231, 24.827]), numpy.array([38.359, 27.084, 19.463])),
            (numpy.array([37.868, 27.017, 24.68]), numpy.array([37.418, 26.49, 19.179])),
            (numpy.array([37.022, 27.047, 25.295]), numpy.array([39.147, 26.168, 19.01])),
            (numpy.array([37.369, 27.619, 25.796]), numpy.array([38.239, 27.371, 19.19])), ]

        for i, frame in enumerate(self.traj):
            ref_coords_first, ref_coords_last = ref_coords[i]
            coords = frame.lipid_coords

            assert_almost_equal(coords[0][0], ref_coords_first, 3)
            assert_almost_equal(coords[-1][-1], ref_coords_last, 3)

    def test_bead_coords(self):
        ref_coords = [(numpy.array([36.704, 26.634, 25.838]), numpy.array([37.928, 25.352, 23.336])),
            (numpy.array([37.328, 25.697, 24.964]), numpy.array([38.159, 25.148, 23.05])),
            (numpy.array([36.654, 25.595, 24.466]), numpy.array([38.012, 24.454, 21.854])),
            (numpy.array([37.523, 26.92, 24.795]), numpy.array([38.869, 25.228, 21.268])),
            (numpy.array([37.233, 25.762, 24.212]), numpy.array([38.983, 24.785, 21.654])),
            (numpy.array([37.778, 26.689, 24.573]), numpy.array([38.953, 24.884, 21.318])),
            (numpy.array([37.344, 27.591, 25.56]), numpy.array([38.506, 25.025, 20.538])),
            (numpy.array([37.307, 27.5, 24.755]), numpy.array([38.814, 25.319, 20.476])),
            (numpy.array([37.68, 27.412, 24.827]), numpy.array([38.366, 24.705, 19.278])),
            (numpy.array([36.885, 27.279, 24.879]), numpy.array([38.489, 24.76, 19.318])),
            (numpy.array([37.049, 27.8, 25.764]), numpy.array([38.63, 25.331, 19.281])), ]

        for i, frame in enumerate(self.traj):
            ref_coords_first, ref_coords_last = ref_coords[i]
            coords = frame.bead_coords

            assert_almost_equal(coords[0], ref_coords_first, 3)
            assert_almost_equal(coords[-1], ref_coords_last, 3)




# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.

from .typedefs cimport real

ctypedef real[2] real_point

# Polygon pseudo class
cdef struct Polygon:
    real_point *points
    int size
    int allocated_size
cdef Polygon *polygon_new() nogil
cdef Polygon *polygon_new_from_memview(real[:, ::1] coords) nogil
cdef Polygon *polygon_new_from_polygon( Polygon *other) nogil
cdef void polygon_copy(Polygon *self, Polygon *other) nogil
cdef void polygon_get_cog(Polygon *self, real_point cog) nogil
cdef void polygon_destroy(Polygon *self) nogil
cdef void polygon_empty(Polygon *self) nogil
cdef real polygon_get_area(Polygon *self) nogil
cdef real polygon_get_perimeter(Polygon *self) nogil
cdef bint polygon_is_inside(Polygon *self, real_point point) nogil
cdef object polygon_as_array(Polygon *self)
cdef void polygon_append(Polygon *self, real_point point) nogil

cdef Polygon *fast_get_clipped_polygon(Polygon *subject, Polygon *clipper) nogil

cdef Polygon *fast_get_zoi(real_point ref_pt, real_point *pts, int size) nogil
cdef void fast_clip_zoi(Polygon *zoi, real_point ref_pt, real_point clipping_pt, Polygon *buffer) nogil




# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function
import sys

__authors__ = [u"Sébastien Buchoux"]
__copyright__ = u"Copyright (c) 2013-2016 %s" % u", ". join(__authors__)
__shortname__ = "FATSLiM"
__cmdlinename__ = "fatslim"
__description__ = "Fast Analysis Toolbox for Simulations of Lipid Membranes"
__full_desc__ = """FATSLiM (Fast Analysis Toolbox for Simulations of Lipid Membranes and its goal
 is to provide an efficient, yet robust, tool to extract physical parameters from MD trajectories.

The main objective of FATSLiM is to decrease as possible the amount of time needed to
analyze MD trajectories: the ultimate goal is to process a several-gigabyte-big file is just a
few minutes or less. This is why a rather important part of FATSLiM's development is
focused on code optimization and simplification in order to maximize its efficiency."""
__licence__ = "GNU Public License 3"
__url__ = "https://github.com/seb-buch/FATSLiM"

version_tuple = (0, 2, 2, 'final', 0)


def _format_version_tuple(version_tup=version_tuple):
    if len(version_tup) == 2:
        main_version = "%d.%d" % version_tup
    else:
        if version_tup[2] == 0:
            main_version = "%d.%d" % version_tup[:2]
        else:
            main_version = "%d.%d.%d" % version_tup[:3]
    if len(version_tup) <= 3:
        return main_version

    release_type = version_tup[3]
    try:
        sub = version_tup[4]
    except IndexError:
        sub = 0

    if release_type == "final":
        return main_version
    elif release_type == "alpha":
        release_type = "a"
    elif release_type == "beta":
        release_type = "b"
    elif release_type == "candidate":
        release_type = "rc"
    else:
        release_type = ".%s" % release_type
    sub_string = "%s%i" % (release_type, sub)
    return main_version + sub_string

__version__ = _format_version_tuple(version_tuple)


def print_greetings():
    print(u"%s - %s" % (__shortname__, __description__))
    print(u"version %s" % __version__)
    print(u"%s\n" % __copyright__)


def print_goodbye(exit_code):
    if exit_code == 0:
        print("Goodbye!")
    else:
        print("Sorry, I made a boo boo...")


def print_full_version_info(with_header=True):
    import platform
    import numpy
    from . import core_base

    if with_header:
        print("%s - %s\n" % (__shortname__, __description__))
    print("%s version: %s" % (__shortname__, __version__))
    print("Python version: %s (%s)" % (platform.python_version(), sys.executable))
    print("Cython version (file generation): %s" % core_base.__cython_version__)
    print("Python compiler: %s" % platform.python_compiler())
    print("CPU architecture: %s" % platform.architecture()[0])
    print("OpenMP: %i CPUs (default number of threads: %i - max: %i)" % (core_base.get_num_procs(),
                                                                         core_base.get_num_threads(),
                                                                         core_base.get_max_threads()))
    print("NumPy version: %s" % numpy.__version__)


_command_registry = None


def main(name=__cmdlinename__, argv=None):
    """Main entry point to FATSLiM user interface

    :param name: Name passed to the CommandRegistry
    :param argv: Command line arguments
    """
    from . import command
    from . import builtins
    import inspect

    try:
        print_greetings()

        global _command_registry
        if _command_registry is None:
            _command_registry = command.CommandRegistry(name)

            for name in dir(builtins):
                if not name.startswith("_"):
                    attr = getattr(builtins, name)
                    if inspect.isclass(attr) and \
                            issubclass(attr, command.Command) and \
                            name.startswith("Cmd"):
                            _command_registry.add_command(attr)
        exit_code = _command_registry.run(argv)
    except KeyboardInterrupt:  # pragma: no cover
        print("Interruption requested!")
        exit_code = 0
    except Exception as exc:  # pragma: no cover
        import traceback
        print("\nFATAL ERROR: %s\n" % exc)
        print("Well, this was unexpected... Congrats, you probably found a bug in %s!" %
              __shortname__)
        print("Please report the following traceback to the devs:\n")
        traceback.print_exc()
        print()
        exit_code = 1

    print_goodbye(exit_code)

    return exit_code


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.

DEF DIM=3

from .typedefs cimport real, rvec, fsl_int

from .core_base cimport Frame
from .core_ns cimport ns_neighborhood_holder

# Classes
cdef class Aggregate:
    # Attributes
    cdef Frame frame
    cdef fsl_int[:] beadids
    cdef fsl_int[:] resids
    cdef fsl_int[:] hg_atomids
    cdef fsl_int[:] lipid_atomids
    cdef real[:, ::1] coords
    cdef real[:, ::1] directions
    cdef real[:, ::1] normals
    cdef real[:, ::1] neighborhood_normals
    cdef real neighborhood_cutoff
    cdef ns_neighborhood_holder *neighborhoods
    cdef rvec xcm
    cdef rvec avg_normal
    cdef bint is_planar
    cdef list lipid_types
    cdef list beadids_by_type
    cdef dict nlipids_by_type

    # Utility
    cdef fsl_int fast_get_atomid(self, fsl_int internalid) nogil
    cpdef fsl_int get_atomid(self, fsl_int internalid)
    cdef bint fast_same_restype(self, fsl_int internalid1, fsl_int internalid2) nogil
    cpdef bint same_restype(self, fsl_int internalid1, fsl_int internalid2)
    cpdef str get_resname(self, fsl_int internalid)
    cpdef fsl_int get_resid(self, fsl_int internalid)

    # APL related
    cdef real apl_cutoff
    cdef real[:] apl_values
    cdef readonly real apl_min
    cdef readonly real apl_max
    cdef readonly real area
    cdef list apl_by_types
    cdef real[:] area_by_types
    cdef readonly real apl_avg

    # Thickness related
    cdef real thickness_cutoff
    cdef real[:] thickness_values
    cdef readonly real thickness_avg
    cdef readonly real thickness_min
    cdef readonly real thickness_max

    # Analysis methods
    cdef void refresh_cache(self) nogil
    cdef fsl_int fast_size(self) nogil
    cdef void compute_neighborhood_normal(self, fsl_int index) nogil
    cdef void set_neighborhood_cutoff(self, real cutoff, bint force=*) nogil
    cdef void compute_apl(self, real cutoff, real area_limit=*, bint force=*) nogil except *
    cdef real fix_apl(self, fsl_int index, real *tmp_apl, bint onlyfix=*) nogil
    cdef void compute_thickness(self, Aggregate other, real interleaflet_cutoff=*, bint force=*) nogil except*
    cdef real fix_thickness(self, fsl_int index, real *tmp_thickness, bint onlyfix=*) nogil


cdef class Membrane:
    # Attributes
    cdef Frame frame
    cdef readonly Aggregate leaflet1
    cdef readonly Aggregate leaflet2
    cdef fsl_int type

    # Thickness related
    cdef real thickness

    # APL related
    cdef real apl

    # Analysis methods
    cdef void fast_compute_thickness(self, real interleaflet_cutoff=*, bint force=*) nogil except*
    cdef void fast_compute_apl(self, real cutoff=*, real area_limit=*, bint force=*) nogil except*



# -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
#  Copyright (C) 2013-2016  Sébastien Buchoux <sebastien.buchoux@gmail.com>
#
#    This file is part of FATSLiM.
#
#    FATSLiM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FATSLiM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FATSLiM.  If not, see <http://www.gnu.org/licenses/>.

import os


def _fullpath_from_basename(fname):
    return os.path.join(DATADIR, fname)

DATADIR = "%s" % os.path.dirname(__file__)

# Model bilayer
MODEL_BILAYER_GRO = _fullpath_from_basename("model_bilayer.gro")
MODEL_BILAYER_NDX = _fullpath_from_basename("model_bilayer.ndx")
MODEL_BILAYER_THICKNESS_XVG = _fullpath_from_basename("model_bilayer_thickness.xvg")
MODEL_BILAYER_THICKNESS_CSV = _fullpath_from_basename("model_bilayer_thickness_raw.csv")
MODEL_BILAYER_APL_CSV = _fullpath_from_basename("model_bilayer_apl_raw.csv")


# Simple (yet experimental) bilayer
BILAYER_GRO = _fullpath_from_basename("bilayer.gro")
BILAYER_NDX = _fullpath_from_basename("bilayer.ndx")

# All-atom bilayer
BILAYER_ALLATOM_GRO = _fullpath_from_basename("bilayer_allatom.gro")
BILAYER_ALLATOM_NDX = _fullpath_from_basename("bilayer_allatom.ndx")

# Simple bilayer with small peptide
BILAYER_PEPTIDE_GRO = _fullpath_from_basename("bilayer_peptide.gro")
BILAYER_PEPTIDE_NDX = _fullpath_from_basename("bilayer_peptide.ndx")
BILAYER_PEPTIDE_H_NDX = _fullpath_from_basename("bilayer_peptide_hydrogen.ndx")

# Bilayer with protein
MODEL_BILAYER_PROT_GRO = _fullpath_from_basename("model_bilayer_prot.gro")
MODEL_BILAYER_PROT_NDX = _fullpath_from_basename("model_bilayer_prot.ndx")
BILAYER_PROT_GRO = _fullpath_from_basename("bilayer_prot.gro")
BILAYER_PROT_NDX = _fullpath_from_basename("bilayer_prot.ndx")
BILAYER_PROT_AREA_XVG = _fullpath_from_basename("bilayer_prot_area.xvg")

# 2 Model bilayers
MODEL_MULTIBILAYER_GRO = _fullpath_from_basename("model_multibilayer.gro")
MODEL_MULTIBILAYER_NDX = _fullpath_from_basename("model_multibilayer.ndx")

# 2 Bilayers
MULTIBILAYER_GRO = _fullpath_from_basename("bilayer_multi.gro")
MULTIBILAYER_NDX = _fullpath_from_basename("bilayer_multi.ndx")

# Model vesicle
MODEL_BILAYER_VESICLE_GRO = _fullpath_from_basename("model_bilayer_vesicle.gro")
MODEL_BILAYER_VESICLE_NDX = _fullpath_from_basename("model_bilayer_vesicle.ndx")

# Bilayer with cholesterol
BILAYER_CHOL_GRO = _fullpath_from_basename("bilayer_chol.gro")
BILAYER_CHOL_NDX = _fullpath_from_basename("bilayer_chol.ndx")
BILAYER_CHOL_APL_GROUPED = _fullpath_from_basename("bilayer_chol_apl_group.xvg")
BILAYER_CHOL_APL_NOGROUP = _fullpath_from_basename("bilayer_chol_apl_nogroup.xvg")

# Big Bilayer-related variables
MODEL_BIG_GRO = _fullpath_from_basename("model_big.gro")
MODEL_BIG_NDX = _fullpath_from_basename("model_big.ndx")

# Big and tricky bilayer
BIG_PROT_GRO = _fullpath_from_basename("deformed.gro")
BIG_PROT_NDX = _fullpath_from_basename("deformed.ndx")

# Vesicle-related variables
MODEL_VESICLE_GRO = _fullpath_from_basename("model_vesicle.gro")
MODEL_VESICLE_NDX = _fullpath_from_basename("model_vesicle.ndx")
VESICLE_GRO = _fullpath_from_basename("dppc_vesicle.gro")
VESICLE_NDX = _fullpath_from_basename("dppc_vesicle.ndx")
VESICLE_XTC = _fullpath_from_basename("dppc_vesicle.xtc")
VESICLE_TRR = _fullpath_from_basename("dppc_vesicle.trr")
VESICLE_APL_XVG = _fullpath_from_basename("dppc_vesicle_apl.xvg")
VESICLE_AREA_XVG = _fullpath_from_basename("dppc_vesicle_area.xvg")
VESICLE_THICKNESS_XVG = _fullpath_from_basename("dppc_vesicle_thickness.xvg")

# Bilayer with ganglioside (i.e. big head groups)
BILAYER_GANGLIO_GRO = _fullpath_from_basename("bilayer_ganglio.gro")
BILAYER_GANGLIO_NDX = _fullpath_from_basename("bilayer_ganglio.ndx")



